# -*- coding: utf-8 -*-
# cython: profile=False
import numpy as np

cimport cython
cimport numpy as np


@cython.cdivision(True)
cdef cdouble u_func(cdouble h2, cdouble w, cdouble e1, double r) nogil:
    return (1 + 1j) * csqrt(-0.5j * (e1 * w ** 2 - h2)) * r
    # return 1j * csqrt(h2 - e1 * w ** 2) * r


@cython.cdivision(True)
cdef cdouble v_func(cdouble h2, cdouble w, cdouble e2, double r) nogil:
    return (1 - 1j) * csqrt(0.5j * (-e2 * w ** 2 + h2)) * r
    # return csqrt(-e2 * w ** 2 + h2) * r


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void eig_mat_with_deriv(
    cdouble h2, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, double ri, cdouble[:, ::1] a, cdouble[:, ::1] b
):
    cdef:
        cdouble vals[3]
        cdouble w2 = w ** 2
        cdouble hew = h2 / e2 / w2
        cdouble ee = e1 / e2
        cdouble x = u_func(h2, w, e1, ri)
        cdouble y = v_func(h2, w, e2, ri)
        cdouble u = u_func(h2, w, e1, r)
        cdouble v = v_func(h2, w, e2, r)
        cdouble ju, jpu, jppu, yu, ypu, yppu, jx, jpx, jppx, yx, ypx, yppx
        cdouble kv, kpv, kppv, iy, ipy, ippy
        cdouble da_du[4][4]
        cdouble da_dv[4][4]
        cdouble da_dx[4][4]
        cdouble da_dy[4][4]
        int sign

    if cimag(u) > 0:
        sign = 1
    elif cimag(u) == 0:
        sign = 0
    else:
        sign = -1

    jve_jvpe_jvppe(n, u, vals)
    ju = vals[0]
    jpu = vals[1]
    jppu = vals[2]
    yve_yvpe_yvppe(n, u, vals)
    yu = vals[0]
    ypu = vals[1]
    yppu = vals[2]
    kve_kvpe_kvppe(n, v, vals)
    kv = vals[0]
    kpv = vals[1]
    kppv = vals[2]
    jve_jvpe_jvppe(n, x, vals)
    jx = vals[0]
    jpx = vals[1]
    jppx = vals[2]
    yve_yvpe_yvppe(n, x, vals)
    yx = vals[0]
    ypx = vals[1]
    yppx = vals[2]
    ive_ivpe_ivppe(n, y, vals)
    iy = vals[0]
    ipy = vals[1]
    ippy = vals[2]

    cdef:
        int i, j
        cdouble du_dh2 = - r ** 2 / (2 * u)
        cdouble dv_dh2 = r ** 2 / (2 * v)
        cdouble dx_dh2 = - ri ** 2 / (2 * x)
        cdouble dy_dh2 = ri ** 2 / (2 * y)
        cdouble nuv = n * (v / u + u / v)
        cdouble dnuv_du = n * (- v / u ** 2 + 1 / v)
        cdouble dnuv_dv = n * (- u / v ** 2 + 1 / u)
        cdouble nxy = n * (y / x + x / y)
        cdouble dnxy_dx = n * (- y / x ** 2 + 1 / y)
        cdouble dnxy_dy = n * (- x / y ** 2 + 1 / x)

    a[0, 0] = jpu * kv * v + kpv * ju * u
    a[0, 1] = ypu * kv * v + kpv * yu * u
    a[0, 2] = nuv * ju * kv
    a[0, 3] = nuv * yu * kv
    a[1, 0] = jpx * y + ipy / iy * jx * x
    a[1, 1] = ypx * y + ipy / iy * yx * x
    a[1, 2] = nxy * jx
    a[1, 3] = nxy * yx
    a[2, 0] = hew * nuv * ju * kv
    a[2, 1] = hew * nuv * yu * kv
    a[2, 2] = ee * jpu * kv * v + kpv * ju * u
    a[2, 3] = ee * ypu * kv * v + kpv * yu * u
    a[3, 0] = hew * nxy * jx
    a[3, 1] = hew * nxy * yx
    a[3, 2] = ee * jpx * y + ipy / iy * jx * x
    a[3, 3] = ee * ypx * y + ipy / iy * yx * x

    da_du[0][0] = jppu * kv * v + kpv * (jpu * u + ju) + 1j * sign * a[0, 0]
    da_du[0][1] = yppu * kv * v + kpv * (ypu * u + yu) + 1j * sign * a[0, 1]
    da_du[0][2] = dnuv_du * ju * kv + nuv * jpu * kv + 1j * sign * a[0, 2]
    da_du[0][3] = dnuv_du * yu * kv + nuv * ypu * kv + 1j * sign * a[0, 3]
    da_du[2][0] = hew * (dnuv_du * ju + nuv * jpu) * kv + 1j * sign * a[2, 0]
    da_du[2][1] = hew * (dnuv_du * yu + nuv * ypu) * kv + 1j * sign * a[2, 1]
    da_du[2][2] = ee * jppu * kv * v + kpv * (jpu * u + ju) + 1j * sign * a[2, 2]
    da_du[2][3] = ee * yppu * kv * v + kpv * (ypu * u + yu) + 1j * sign * a[2, 3]
    for i in range(4):
        da_du[1][i] = da_du[3][i] = 0.0

    da_dv[0][0] = jpu * (kpv * v + kv) + kppv * ju * u + a[0, 0]
    da_dv[0][1] = ypu * (kpv * v + kv) + kppv * yu * u + a[0, 1]
    da_dv[0][2] = (dnuv_dv * kv + nuv * kpv) * ju + a[0, 2]
    da_dv[0][3] = (dnuv_dv * kv + nuv * kpv) * yu + a[0, 3]
    for i in range(4):
        da_dv[1][i] = da_dv[3][i] = 0.0
    da_dv[2][0] = hew * (dnuv_dv * kv + nuv * kpv) * ju + a[2, 0]
    da_dv[2][1] = hew * (dnuv_dv * kv + nuv * kpv) * yu + a[2, 1]
    da_dv[2][2] = ee * jpu * (kpv * v + kv) + kppv * ju * u + a[2, 2]
    da_dv[2][3] = ee * ypu * (kpv * v + kv) + kppv * yu * u + a[2, 3]

    for i in range(4):
        da_dx[0][i] = da_dx[2][i] = 0.0
    da_dx[1][0] = jppx * y + ipy / iy * (jpx * x + jx) + 1j * sign * a[1, 0]
    da_dx[1][1] = yppx * y + ipy / iy * (ypx * x + yx) + 1j * sign * a[1, 1]
    da_dx[1][2] = dnxy_dx * jx + nxy * jpx + 1j * sign * a[1, 2]
    da_dx[1][3] = dnxy_dx * yx + nxy * ypx + 1j * sign * a[1, 3]
    da_dx[3][0] = hew * (dnxy_dx * jx + nxy * jpx) + 1j * sign * a[3, 0]
    da_dx[3][1] = hew * (dnxy_dx * yx + nxy * ypx) + 1j * sign * a[3, 1]
    da_dx[3][2] = ee * jppx * y + ipy / iy * (jpx * x + jx) + 1j * sign * a[3, 2]
    da_dx[3][3] = ee * yppx * y + ipy / iy * (ypx * x + yx) + 1j * sign * a[3, 3]

    for i in range(4):
        da_dy[0][i] = da_dy[2][i] = 0.0
    da_dy[1][0] = jpx + (ippy / iy - ipy ** 2 / iy ** 2) * jx * x
    da_dy[1][1] = ypx + (ippy / iy - ipy ** 2 / iy ** 2) * yx * x
    da_dy[1][2] = dnxy_dy * jx
    da_dy[1][3] = dnxy_dy * yx
    da_dy[3][0] = hew * dnxy_dy * jx
    da_dy[3][1] = hew * dnxy_dy * yx
    da_dy[3][2] = ee * jpx + (ippy / iy - ipy ** 2 / iy ** 2) * jx * x
    da_dy[3][3] = ee * ypx + (ippy / iy - ipy ** 2 / iy ** 2) * yx * x

    for i in range(4):
        for j in range(4):
            b[i, j] = (da_du[i][j] * du_dh2 + da_dv[i][j] * dv_dh2
                        + da_dx[i][j] * dx_dh2 + da_dy[i][j] * dy_dh2)
    b[2, 0] += 1.0 / (e2 * w2) * nuv * ju * kv
    b[2, 1] += 1.0 / (e2 * w2) * nuv * yu * kv
    b[3, 0] += 1.0 / (e2 * w2) * nxy * jx
    b[3, 1] += 1.0 / (e2 * w2) * nxy * yx


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def eig_eq_with_jac(
    double[::1] h2vec, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, double ri, cdouble[::1] roots
) -> tuple[np.ndarray, np.ndarray]:
    """Return the value of the characteristic equation

    Args:
        h2vec: The real and imaginary parts of the square of propagation constant.
        w: The angular frequency
        pol: The polarization
        n: The order of the modes
        e1: The permittivity of the core
        e2: The permittivity of the clad.
    Returns:
        val: A complex indicating the left-hand value of the characteristic
            equation.
    """
    cdef:
        cdouble h2 = h2vec[0] + h2vec[1] * 1j
        cdouble[:, ::1] a = np.empty((4, 4), dtype=complex)
        cdouble[:, ::1] b = np.empty((4, 4), dtype=complex)
        cdouble f, fp, dd, ddi, denom
        int i
        int num = len(roots)

    eig_mat_with_deriv(h2, w, pol, n, e1, e2, r, ri, a, b)
    if n == 0:
        if pol == "E":
            f = a[0, 0] * a[1, 1] - a[0, 1] * a[1, 0]
            fp = deriv_det2(
                np.ascontiguousarray(a[:2, :2]),
                np.ascontiguousarray(b[:2, :2]))
        else:
            f = a[2, 2] * a[3, 3] - a[2, 3] * a[3, 2]
            fp = deriv_det2(
                np.ascontiguousarray(a[2:, 2:]),
                np.ascontiguousarray(b[2:, 2:]))
    else:
        f = det4(a)
        fp = deriv_det4(a, b)
    denom = 1.0
    dd = 0.0
    for i in range(num):
        denom *= (h2 - roots[i]) / roots[i]
        ddi = - roots[i] / (h2 - roots[i]) ** 2
        for j in range(num):
            if j != i:
                ddi /= (h2 - roots[j]) / roots[j]
        dd += ddi
    fp = fp / denom + f * dd
    f /= denom
    return np.array([f.real, f.imag]), np.array([[fp.real, fp.imag], [-fp.imag, fp.real]])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def eig_eq_for_min_with_jac(
    double[::1] h2vec, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, double ri, cdouble[::1] roots
) -> tuple[float, np.ndarray]:
    """Return the value of the characteristic equation

    Args:
        h2vec: The real and imaginary parts of the square of propagation constant.
        w: The angular frequency
        pol: The polarization
        n: The order of the modes
        e1: The permittivity of the core
        e2: The permittivity of the clad.
    Returns:
        val: A complex indicating the left-hand value of the characteristic
            equation.
    """
    cdef:
        cdouble h2 = h2vec[0] + h2vec[1] * 1j
        cdouble[:, ::1] a = np.empty((4, 4), dtype=complex)
        cdouble[:, ::1] b = np.empty((4, 4), dtype=complex)
        cdouble f, val, fp, dd, ddi, denom
        int i, j
        int num = len(roots)

    eig_mat_with_deriv(h2, w, pol, n, e1, e2, r, ri, a, b)
    if n == 0:
        if pol == "E":
            f = a[0, 0] * a[1, 1] - a[0, 1] * a[1, 0]
            fp = deriv_det2(
                np.ascontiguousarray(a[:2, :2]),
                np.ascontiguousarray(b[:2, :2]))
        else:
            f = a[2, 2] * a[3, 3] - a[2, 3] * a[3, 2]
            fp = deriv_det2(
                np.ascontiguousarray(a[2:, 2:]),
                np.ascontiguousarray(b[2:, 2:]))
    else:
        f = det4(a)
        fp = deriv_det4(a, b)
    denom = 1.0
    dd = 0.0
    for i in range(num):
        denom *= (h2 - roots[i]) / roots[i]
        ddi = - roots[i] / (h2 - roots[i]) ** 2
        for j in range(num):
            if j != i:
                ddi /= (h2 - roots[j]) / roots[j]
        dd += ddi
    fp = fp / denom + f * dd
    f /= denom
    return (f.real ** 2 + f.imag ** 2,
            np.array([
                2 * (f.real * fp.real + f.imag * fp.imag),
                2 * (-f.real * fp.imag + f.imag * fp.real)]))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void eig_mat(
    cdouble h2, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, double ri, cdouble[:, ::1] a
):
    cdef:
        cdouble vals[2]
        cdouble w2 = w ** 2
        cdouble hew = h2 / e2 / w2
        cdouble ee = e1 / e2
        cdouble x = u_func(h2, w, e1, ri)
        cdouble y = v_func(h2, w, e2, ri)
        cdouble u = u_func(h2, w, e1, r)
        cdouble v = v_func(h2, w, e2, r)
        cdouble ju, jpu, yu, ypu, jx, jpx, yx, ypx
        cdouble kv, kpv, iy, ipy
        cdouble nuv = n * (v / u + u / v)
        cdouble nxy = n * (y / x + x / y)

    jve_jvpe(n, u, vals)
    ju = vals[0]
    jpu = vals[1]
    yve_yvpe(n, u, vals)
    yu = vals[0]
    ypu = vals[1]
    kve_kvpe(n, v, vals)
    kv = vals[0]
    kpv = vals[1]
    jve_jvpe(n, x, vals)
    jx = vals[0]
    jpx = vals[1]
    yve_yvpe(n, x, vals)
    yx = vals[0]
    ypx = vals[1]
    ive_ivpe(n, y, vals)
    iy = vals[0]
    ipy = vals[1]

    a[0, 0] = jpu * kv * v + kpv * ju * u
    a[0, 1] = ypu * kv * v + kpv * yu * u
    a[0, 2] = nuv * ju * kv
    a[0, 3] = nuv * yu * kv
    a[1, 0] = jpx / yx * y + ipy / iy * jx / yx * x
    a[1, 1] = ypx / yx * y + ipy / iy * x
    a[1, 2] = nxy * jx / yx
    a[1, 3] = nxy
    a[2, 0] = hew * nuv * ju * kv
    a[2, 1] = hew * nuv * yu * kv
    a[2, 2] = ee * jpu * kv * v + kpv * ju * u
    a[2, 3] = ee * ypu * kv * v + kpv * yu * u
    a[3, 0] = hew * nxy * jx / yx
    a[3, 1] = hew * nxy
    a[3, 2] = ee * jpx / yx * y + ipy / iy * jx / yx * x
    a[3, 3] = ee * ypx / yx * y + ipy / iy * x


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def eig_eq(
    double[::1] h2vec, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, double ri, cdouble[::1] roots
) -> np.ndarray:
    """Return the value of the characteristic equation

    Args:
        h2vec: The real and imaginary parts of the square of propagation constant.
        w: The angular frequency
        pol: The polarization
        n: The order of the modes
        e1: The permittivity of the core
        e2: The permittivity of the clad.
    Returns:
        np.array([f.real, f.imag]):  Real and imaginary parts of left-hand value of the characteristic equation.
    """
    cdef:
        cdouble h2 = h2vec[0] + h2vec[1] * 1j
        cdouble[:, ::1] a = np.empty((4, 4), dtype=complex)
        cdouble f, fp, denom
        int i
        int num = len(roots)

    eig_mat(h2, w, pol, n, e1, e2, r, ri, a)
    if n == 0:
        if pol == "E":
            f = a[0, 0] * a[1, 1] - a[0, 1] * a[1, 0]
        else:
            f = a[2, 2] * a[3, 3] - a[2, 3] * a[3, 2]
    else:
        f = det4(a)
    denom = 1.0
    for i in range(num):
        denom *= (h2 - roots[i]) / roots[i]
    f /= denom
    return np.array([f.real, f.imag])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def eig_eq_for_min(
    double[::1] h2vec, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, double ri, cdouble[::1] roots
) -> tuple[float, np.ndarray]:
    """Return the value of the characteristic equation

    Args:
        h2vec: The real and imaginary parts of the square of propagation constant.
        w: The angular frequency
        pol: The polarization
        n: The order of the modes
        e1: The permittivity of the core
        e2: The permittivity of the clad.
    Returns:
        val: A complex indicating the left-hand value of the characteristic
            equation.
    """
    cdef:
        cdouble h2 = h2vec[0] + h2vec[1] * 1j
        cdouble[:, ::1] a = np.empty((4, 4), dtype=complex)
        cdouble f, val, dd, ddi, denom
        int i, j
        int num = len(roots)

    eig_mat(h2, w, pol, n, e1, e2, r, ri, a)
    if n == 0:
        if pol == "E":
            f = a[0, 0] * a[1, 1] - a[0, 1] * a[1, 0]
        else:
            f = a[2, 2] * a[3, 3] - a[2, 3] * a[3, 2]
    else:
        f = det4(a)
    denom = 1.0
    for i in range(num):
        denom *= (h2 - roots[i]) / roots[i]
    f /= denom
    return log10(f.real ** 2 + f.imag ** 2)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void eig_mat_u(
    cdouble u, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, double ri, cdouble[:, ::1] a
):
    cdef:
        cdouble vals[2]
        cdouble w2 = w ** 2
        cdouble h2 = e1 * w2 - u ** 2 / r ** 2
        cdouble hew = h2 / e2 / w2
        cdouble ee = e1 / e2
        cdouble x = u * ri / r
        cdouble y = v_func(h2, w, e2, ri)
        cdouble v = v_func(h2, w, e2, r)
        cdouble ju, jpu, yu, ypu, jx, jpx, yx, ypx
        cdouble kv, kpv, iy, ipy
        cdouble nuv = n * (v / u + u / v)
        cdouble nxy = n * (y / x + x / y)

    jve_jvpe(n, u, vals)
    ju = vals[0]
    jpu = vals[1]
    yve_yvpe(n, u, vals)
    yu = vals[0]
    ypu = vals[1]
    kve_kvpe(n, v, vals)
    kv = vals[0]
    kpv = vals[1]
    jve_jvpe(n, x, vals)
    jx = vals[0]
    jpx = vals[1]
    yve_yvpe(n, x, vals)
    yx = vals[0]
    ypx = vals[1]
    ive_ivpe(n, y, vals)
    iy = vals[0]
    ipy = vals[1]

    a[0, 0] = jpu * kv * v + kpv * ju * u
    a[0, 1] = ypu * kv * v + kpv * yu * u
    a[0, 2] = nuv * ju * kv
    a[0, 3] = nuv * yu * kv
    a[1, 0] = jpx / yx * y + ipy / iy * jx / yx * x
    a[1, 1] = ypx / yx * y + ipy / iy * x
    a[1, 2] = nxy * jx / yx
    a[1, 3] = nxy
    a[2, 0] = hew * nuv * ju * kv
    a[2, 1] = hew * nuv * yu * kv
    a[2, 2] = ee * jpu * kv * v + kpv * ju * u
    a[2, 3] = ee * ypu * kv * v + kpv * yu * u
    a[3, 0] = hew * nxy * jx / yx
    a[3, 1] = hew * nxy
    a[3, 2] = ee * jpx / yx * y + ipy / iy * jx / yx * x
    a[3, 3] = ee * ypx / yx * y + ipy / iy * x


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def eig_eq_u(
    double[::1] uvec, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, double ri, cdouble[::1] roots
) -> np.ndarray:
    """Return the value of the characteristic equation

    Args:
        uvec: The real and imaginary parts of u.
        w: The angular frequency
        pol: The polarization
        n: The order of the modes
        e1: The permittivity of the core
        e2: The permittivity of the clad.
    Returns:
        f.real, f.imag:  The left-hand value of the characteristic equation.
    """
    cdef:
        cdouble u = uvec[0] + uvec[1] * 1j
        # cdouble hu = hankel2(n, u)
        cdouble[:, ::1] a = np.empty((4, 4), dtype=complex)
        cdouble f, denom
        int i
        int num = len(roots)

    eig_mat_u(u, w, pol, n, e1, e2, r, ri, a)
    if n == 0:
        if pol == "E":
            f = a[0, 0] * a[1, 1] - a[0, 1] * a[1, 0]
        else:
            f = a[2, 2] * a[3, 3] - a[2, 3] * a[3, 2]
    else:
        f = det4(a)
    # denom = hu
    denom = 1.0
    for i in range(num):
        denom *= (u - roots[i]) / roots[i]
    f /= denom
    return np.array([f.real, f.imag])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def eig_eq_u_for_min(
    double[::1] uvec, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, double ri, cdouble[::1] roots
) -> double:
    """Return the value of the characteristic equation

    Args:
        uvec: The real and imaginary parts of u.
        w: The angular frequency
        pol: The polarization
        n: The order of the modes
        e1: The permittivity of the core
        e2: The permittivity of the clad.
    Returns:
        f:  The left-hand value of the characteristic equation.
    """
    cdef:
        cdouble u = uvec[0] + uvec[1] * 1j
        # cdouble hu = hankel2(n, u)
        cdouble[:, ::1] a = np.empty((4, 4), dtype=complex)
        cdouble f, denom
        int i
        int num = len(roots)

    eig_mat_u(u, w, pol, n, e1, e2, r, ri, a)
    if n == 0:
        if pol == "E":
            f = a[0, 0] * a[1, 1] - a[0, 1] * a[1, 0]
        else:
            f = a[2, 2] * a[3, 3] - a[2, 3] * a[3, 2]
    else:
        f = det4(a)
    # denom = hu
    denom = 1.0
    for i in range(num):
        denom *= (u - roots[i]) / roots[i]
    f /= denom
    return cabs(f)
    # return log10(cabs(f))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void eig_mat_u_with_deriv(
    cdouble u, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, double ri, cdouble[:, ::1] a, cdouble[:, ::1] b
):
    cdef:
        cdouble vals[3]
        cdouble w2 = w ** 2
        cdouble h2 = e1 * w2 - u ** 2 / r ** 2
        cdouble hew = h2 / e2 / w2
        cdouble ee = e1 / e2
        cdouble x = u * ri / r
        cdouble y = v_func(h2, w, e2, ri)
        cdouble v = v_func(h2, w, e2, r)
        cdouble ju, jpu, jppu, yu, ypu, yppu, jx, jpx, jppx, yx, ypx, yppx
        cdouble kv, kpv, kppv, iy, ipy, ippy
        cdouble da_du[4][4]
        cdouble da_dv[4][4]
        cdouble da_dx[4][4]
        cdouble da_dy[4][4]
        int sign

    if cimag(u) > 0:
        sign = 1
    elif cimag(u) == 0:
        sign = 0
    else:
        sign = -1

    jve_jvpe_jvppe(n, u, vals)
    ju = vals[0]
    jpu = vals[1]
    jppu = vals[2]
    yve_yvpe_yvppe(n, u, vals)
    yu = vals[0]
    ypu = vals[1]
    yppu = vals[2]
    kve_kvpe_kvppe(n, v, vals)
    kv = vals[0]
    kpv = vals[1]
    kppv = vals[2]
    jve_jvpe_jvppe(n, x, vals)
    jx = vals[0]
    jpx = vals[1]
    jppx = vals[2]
    yve_yvpe_yvppe(n, x, vals)
    yx = vals[0]
    ypx = vals[1]
    yppx = vals[2]
    ive_ivpe_ivppe(n, y, vals)
    iy = vals[0]
    ipy = vals[1]
    ippy = vals[2]

    cdef:
        int i, j
        cdouble dh2_du = - 2 * u / r ** 2
        cdouble dv_du = - u / v
        cdouble dx_du = ri / r
        cdouble dy_du = - x / v

        cdouble nuv = n * (v / u + u / v)
        cdouble dnuv_du = n * (- v / u ** 2 + 1 / v)
        cdouble dnuv_dv = n * (- u / v ** 2 + 1 / u)
        cdouble nxy = n * (y / x + x / y)
        cdouble dnxy_dx = n * (- y / x ** 2 + 1 / y)
        cdouble dnxy_dy = n * (- x / y ** 2 + 1 / x)

    a[0, 0] = jpu * kv * v + kpv * ju * u
    a[0, 1] = ypu * kv * v + kpv * yu * u
    a[0, 2] = nuv * ju * kv
    a[0, 3] = nuv * yu * kv
    a[1, 0] = jpx / yx * y + ipy / iy * jx / yx * x
    a[1, 1] = ypx / yx * y + ipy / iy * x
    a[1, 2] = nxy * jx / yx
    a[1, 3] = nxy
    a[2, 0] = hew * nuv * ju * kv
    a[2, 1] = hew * nuv * yu * kv
    a[2, 2] = ee * jpu * kv * v + kpv * ju * u
    a[2, 3] = ee * ypu * kv * v + kpv * yu * u
    a[3, 0] = hew * nxy * jx / yx
    a[3, 1] = hew * nxy
    a[3, 2] = ee * jpx / yx * y + ipy / iy * jx / yx * x
    a[3, 3] = ee * ypx / yx * y + ipy / iy * x

    da_du[0][0] = jppu * kv * v + kpv * (jpu * u + ju) + 1j * sign * a[0, 0]
    da_du[0][1] = yppu * kv * v + kpv * (ypu * u + yu) + 1j * sign * a[0, 1]
    da_du[0][2] = dnuv_du * ju * kv + nuv * jpu * kv + 1j * sign * a[0, 2]
    da_du[0][3] = dnuv_du * yu * kv + nuv * ypu * kv + 1j * sign * a[0, 3]
    da_du[2][0] = hew * (dnuv_du * ju + nuv * jpu) * kv + 1j * sign * a[2, 0]
    da_du[2][1] = hew * (dnuv_du * yu + nuv * ypu) * kv + 1j * sign * a[2, 1]
    da_du[2][2] = ee * jppu * kv * v + kpv * (jpu * u + ju) + 1j * sign * a[2, 2]
    da_du[2][3] = ee * yppu * kv * v + kpv * (ypu * u + yu) + 1j * sign * a[2, 3]
    for i in range(4):
        da_du[1][i] = da_du[3][i] = 0.0

    da_dv[0][0] = jpu * (kpv * v + kv) + kppv * ju * u + a[0, 0]
    da_dv[0][1] = ypu * (kpv * v + kv) + kppv * yu * u + a[0, 1]
    da_dv[0][2] = (dnuv_dv * kv + nuv * kpv) * ju + a[0, 2]
    da_dv[0][3] = (dnuv_dv * kv + nuv * kpv) * yu + a[0, 3]
    for i in range(4):
        da_dv[1][i] = da_dv[3][i] = 0.0
    da_dv[2][0] = hew * (dnuv_dv * kv + nuv * kpv) * ju + a[2, 0]
    da_dv[2][1] = hew * (dnuv_dv * kv + nuv * kpv) * yu + a[2, 1]
    da_dv[2][2] = ee * jpu * (kpv * v + kv) + kppv * ju * u + a[2, 2]
    da_dv[2][3] = ee * ypu * (kpv * v + kv) + kppv * yu * u + a[2, 3]

    for i in range(4):
        da_dx[0][i] = da_dx[2][i] = 0.0
    da_dx[1][0] = ((jppx / yx - jpx * ypx / yx ** 2) * y
        + ipy / iy * ((jpx / yx - jx * ypx / yx ** 2) * x + jx / yx))
    da_dx[1][1] = (yppx / yx - ypx ** 2 / yx ** 2) * y + ipy / iy
    da_dx[1][2] = dnxy_dx * jx / yx + nxy * jpx / yx - nxy * jx * ypx / yx ** 2
    da_dx[1][3] = dnxy_dx
    da_dx[3][0] = hew * (dnxy_dx * jx / yx + nxy * jpx / yx - nxy * jx * ypx / yx ** 2)
    da_dx[3][1] = hew * dnxy_dx
    da_dx[3][2] = (ee * (jppx / yx - jpx * ypx / yx ** 2) * y
        + ipy / iy * ((jpx / yx - jx * ypx / yx ** 2) * x + jx / yx))
    da_dx[3][3] = ee * (yppx / yx - ypx ** 2 / yx ** 2) * y + ipy / iy

    for i in range(4):
        da_dy[0][i] = da_dy[2][i] = 0.0
    da_dy[1][0] = jpx / yx + (ippy / iy - ipy ** 2 / iy ** 2) * jx / yx * x
    da_dy[1][1] = ypx / yx + (ippy / iy - ipy ** 2 / iy ** 2) * x
    da_dy[1][2] = dnxy_dy * jx / yx
    da_dy[1][3] = dnxy_dy
    da_dy[3][0] = hew * dnxy_dy * jx / yx
    da_dy[3][1] = hew * dnxy_dy
    da_dy[3][2] = ee * jpx / yx + (ippy / iy - ipy ** 2 / iy ** 2) * jx / yx * x
    da_dy[3][3] = ee * ypx / yx + (ippy / iy - ipy ** 2 / iy ** 2) * x

    for i in range(4):
        for j in range(4):
            b[i, j] = (da_du[i][j] + da_dv[i][j] * dv_du
                        + da_dx[i][j] * dx_du + da_dy[i][j] * dy_du)
    b[2, 0] += 1.0 / (e2 * w2) * nuv * ju * kv * dh2_du
    b[2, 1] += 1.0 / (e2 * w2) * nuv * yu * kv * dh2_du
    b[3, 0] += 1.0 / (e2 * w2) * nxy * jx / yx * dh2_du
    b[3, 1] += 1.0 / (e2 * w2) * nxy * dh2_du


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def eig_eq_u_with_jac(
    double[::1] uvec, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, double ri, cdouble[::1] roots
) -> tuple[np.ndarray, np.ndarray]:
    """Return the value of the characteristic equation

    Args:
        uvec: The real and imaginary parts of u.
        w: The angular frequency
        pol: The polarization
        n: The order of the modes
        e1: The permittivity of the core
        e2: The permittivity of the clad.
    Returns:
        val: A complex indicating the left-hand value of the characteristic
            equation.
    """
    cdef:
        cdouble u = uvec[0] + uvec[1] * 1j
        cdouble[:, ::1] a = np.empty((4, 4), dtype=complex)
        cdouble[:, ::1] b = np.empty((4, 4), dtype=complex)
        cdouble f, fp, dd, ddi, denom
        int i
        int num = len(roots)

    eig_mat_u_with_deriv(u, w, pol, n, e1, e2, r, ri, a, b)
    if n == 0:
        if pol == "E":
            f = a[0, 0] * a[1, 1] - a[0, 1] * a[1, 0]
            fp = deriv_det2(
                np.ascontiguousarray(a[:2, :2]),
                np.ascontiguousarray(b[:2, :2]))
        else:
            f = a[2, 2] * a[3, 3] - a[2, 3] * a[3, 2]
            fp = deriv_det2(
                np.ascontiguousarray(a[2:, 2:]),
                np.ascontiguousarray(b[2:, 2:]))
    else:
        f = det4(a)
        fp = deriv_det4(a, b)
    denom = 1.0
    dd = 0.0
    for i in range(num):
        denom *= (u - roots[i]) / roots[i]
        ddi = - roots[i] / (u - roots[i]) ** 2
        for j in range(num):
            if j != i:
                ddi /= (u - roots[j]) / roots[j]
        dd += ddi
    fp = fp / denom + f * dd
    f /= denom
    return np.array([f.real, f.imag]), np.array([[fp.real, fp.imag], [-fp.imag, fp.real]])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void coefs_C(
    cdouble *hs, cdouble w, long *s_all, long *n_all,
    int num_n_all, double r, double ri, cdouble e1, cdouble e2,
    cdouble *xs, cdouble *ys, cdouble *us, cdouble *vs,
    cdouble *jxs, cdouble *jpxs, cdouble *yxs, cdouble *ypxs, cdouble *iys, cdouble *ipys,
    cdouble *jus, cdouble *jpus, cdouble *yus, cdouble *ypus, cdouble *kvs, cdouble *kpvs,
    cdouble *A1s, cdouble *B1s, cdouble *A2s, cdouble *B2s,
    cdouble *C2s, cdouble *D2s, cdouble *A3s, cdouble *B3s,
    cdouble *Ys) nogil:
    cdef:
        int i, s, n, en
        cdouble ee = e1 / e2
        cdouble val
        cdouble I1, I2, I3, Y1, Y2, Y3
        cdouble norm, Y, hew
        cdouble h, a1, b1, a2, b2, c2, d2, a3, b3
        cdouble x, y, u, v
        cdouble ju, jpu, yu, ypu, jx, jpx, yx, ypx
        cdouble kv, kpv, iy, ipy, nuv, nxy
        cdouble y_te, y_tm1, y_tm2
        cdouble A[3][3]
        cdouble B[3]
        cdouble C[3]
        cdouble vals[2]
    for i in range(num_n_all):
        n = n_all[i]
        en = 1 if n == 0 else 2
        h = hs[i]
        y_te = h / w
        y_tm1 = e1 * w / h
        y_tm2 = e2 * w / h
        s = s_all[i]
        u = u_func(h * h, w, e1, r)
        jv_jvp(n, u, vals)
        ju = vals[0]
        jpu = vals[1]
        yv_yvp(n, u, vals)
        yu = vals[0]
        ypu = vals[1]
        x = u_func(h * h, w, e1, ri)
        jv_jvp(n, x, vals)
        jx = vals[0]
        jpx = vals[1]
        yv_yvp(n, x, vals)
        yx = vals[0]
        ypx = vals[1]

        if creal(e2) < -1e6:
            a1 = b1 = a3 = b3 = 0.0j
            if s == 0:
                a2 = 1.0 + 0.0j
                c2 = - jpu / ypu
                b2 = d2 = 0.0j
            else:
                b2 = 1.0 + 0.0j
                d2 = - ju / yu
                a2 = c2 = 0.0j
            y = INFINITY
            v = INFINITY
            iys[i] = INFINITY
            ipys[i] = INFINITY
            kvs[i] = 0.0
            kpvs[i] = 0.0
        else:
            hew = h ** 2 / e2 / w ** 2
            y = v_func(h * h, w, e2, ri)
            v = v_func(h * h, w, e2, r)
            kv_kvp(n, v, vals)
            kv = vals[0]
            kpv = vals[1]
            iv_ivp(n, y, vals)
            iy = vals[0]
            ipy = vals[1]
            nuv = n * (v / u + u / v)
            nxy = n * (y / x + x / y)
            if s == 0:
                A[0][0] = ypx / yx * y + ipy / iy * x
                A[0][1] = nxy * jx / yx
                A[0][2] = nxy
                A[1][0] = hew * nuv * yu * kv
                A[1][1] = ee * jpu * kv * v + kpv * ju * u
                A[1][2] = ee * ypu * kv * v + kpv * yu * u
                A[2][0] = hew * nxy
                A[2][1] = ee * jpx / yx * y + ipy / iy * jx / yx * x
                A[2][2] = ee * ypx / yx * y + ipy / iy * x
                B[0] = - (jpx / yx * y + ipy / iy * jx / yx * x)
                B[1] = - (hew * nuv * ju * kv)
                B[2] = - (hew * nxy * jx / yx)
                a2 = 1.0 + 0j
                solve(A, B, C)
                c2 = C[0]
                b2 = C[1]
                d2 = C[2]
            else:
                A[0][0] = ee * ypx / yx * y + ipy / iy * x
                A[0][1] = hew * nxy * jx / yx
                A[0][2] = hew * nxy
                A[1][0] = nuv * yu * kv
                A[1][1] = jpu * kv * v + kpv * ju * u
                A[1][2] = ypu * kv * v + kpv * yu * u
                A[2][0] = nxy
                A[2][1] = jpx / yx * y + ipy / iy * jx / yx * x
                A[2][2] = ypx / yx * y + ipy / iy * x
                B[0] = - (ee * jpx / yx * y + ipy / iy * jx / yx * x)
                B[1] = - (nuv * ju * kv)
                B[2] = - (nxy * jx / yx)
                b2 = 1.0 + 0j
                solve(A, B, C)
                d2 = C[0]
                a2 = C[1]
                c2 = C[2]
            a1 = - x / (y * iy) * (jx * a2 + yx * c2)
            b1 = - x / (y * iy) * (jx * b2 + yx * d2)
            a3 = - u / (v * kv) * (ju * a2 + yu * c2)
            b3 = - u / (v * kv) * (ju * b2 + yu * d2)

        val = M_PI / en * (
            r ** 2 * (jpu ** 2 + (1 - n ** 2 / u ** 2) * ju ** 2 + 2 * jpu * ju / u)
            - ri ** 2 * (jpx ** 2 + (1 - n ** 2 / x ** 2) * jx ** 2 + 2 * jpx * jx / x)
        )
        I2 = val * (a2 ** 2 + b2 ** 2)
        Y2 = val * (y_te * a2 ** 2 + y_tm1 * b2 ** 2)
        val = M_PI / en * (
            r ** 2 * (ypu ** 2 + (1 - n ** 2 / u ** 2) * yu ** 2 + 2 * ypu * yu / u)
            - ri ** 2 * (ypx ** 2 + (1 - n ** 2 / x ** 2) * yx ** 2 + 2 * ypx * yx / x)
        )
        I2 += val * (c2 ** 2 + d2 ** 2)
        Y2 += val * (y_te * c2 ** 2 + y_tm1 * d2 ** 2)
        val = M_PI / en * 2 * (
            r ** 2 * (jpu * ypu + (1 - n ** 2 / u ** 2) * ju * yu + 2 * jpu * yu / u)
            - ri ** 2 * (jpx * ypx + (1 - n ** 2 / x ** 2) * jx * yx + 2 * jpx * yx / x)
        )
        I2 += val * (a2 * c2 + b2 * d2)
        Y2 += val * (y_te * a2 * c2 + y_tm1 * b2 * d2)
        val = M_PI * n * (r ** 2 / u ** 2 * ju ** 2 - ri ** 2 / x ** 2 * jx ** 2)
        I2 += val * 2 * a2 * b2
        Y2 += val * (y_te + y_tm1) * a2 * b2
        val = M_PI * n * (r ** 2 / u ** 2 * yu ** 2 - ri ** 2 / x ** 2 * yx ** 2)
        I2 += 2 * val * c2 * d2
        Y2 += val * (y_te + y_tm1) * c2 * d2
        val = M_PI * n *  (r ** 2 / u ** 2 * ju * yu - ri ** 2 / x ** 2 * jx * yx)
        I2 += 2 * val * (a2 * d2 + b2 * c2)
        Y2 +=  val * (y_te + y_tm1) * (a2 * d2 + b2 * c2)

        if creal(e2) < -1e6:
            norm = csqrt(I2)
            Y = Y2 / I2
        else:
            val = M_PI * ri ** 2 / en * (ipy ** 2 - (1 + n ** 2 / y ** 2) * iy ** 2 + 2 * ipy * iy / y)
            I1 = val * (a1 ** 2 + b1 ** 2)
            Y1 = val * (y_te * a1 ** 2 + y_tm2 * b1 ** 2)
            val = M_PI * ri ** 2 * n * iy ** 2 / y ** 2
            I1 += val * 2 * a1 * b1
            Y1 += val * (y_te + y_tm2) * a1 * b1
            val = - M_PI * r ** 2 / en * (kpv ** 2 - (1 + n ** 2 / v ** 2) * kv ** 2 + 2 * kpv * kv / v)
            I3 = val * (a3 ** 2 + b3 ** 2)
            Y3 = val * (y_te * a3 ** 2 + y_tm2 * b3 ** 2)
            val = - M_PI * r ** 2 * n * kv ** 2 / v ** 2
            I3 += val * 2 * a3 * b3
            Y3 += val * (y_te + y_tm2) * a3 * b3
            norm = csqrt(I1 + I2 + I3)
            Y = (Y1 + Y2 + Y3) / (I1 + I2 + I3)

        xs[i] = x
        ys[i] = y
        us[i] = u
        vs[i] = v
        jxs[i] = jx
        jpxs[i] = jpx
        yxs[i] = yx
        ypxs[i] = ypx
        iys [i] = iy
        ipys[i] = ipy
        jus[i] = ju
        jpus[i] = jpu
        yus[i] = yu
        ypus[i] = ypu
        kvs [i] = kv
        kpvs[i] = kpv
        A1s[i] = a1 / norm
        B1s[i] = b1 / norm
        A2s[i] = a2 / norm
        B2s[i] = b2 / norm
        C2s[i] = c2 / norm
        D2s[i] = d2 / norm
        A3s[i] = a3 / norm
        B3s[i] = b3 / norm
        Ys[i] = Y


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def props_cython(cdouble w, double r, double ri, long[::1] s_all, long[::1] n_all,
                 cdouble[::1] hs, cdouble e1, cdouble e2):
    cdef:
        int num_n_all = n_all.shape[0]
    xs_array = np.empty(num_n_all, dtype=complex)
    ys_array = np.empty(num_n_all, dtype=complex)
    us_array = np.empty(num_n_all, dtype=complex)
    vs_array = np.empty(num_n_all, dtype=complex)
    jxs_array = np.empty(num_n_all, dtype=complex)
    jpxs_array = np.empty(num_n_all, dtype=complex)
    yxs_array = np.empty(num_n_all, dtype=complex)
    ypxs_array = np.empty(num_n_all, dtype=complex)
    iys_array = np.empty(num_n_all, dtype=complex)
    ipys_array = np.empty(num_n_all, dtype=complex)
    jus_array = np.empty(num_n_all, dtype=complex)
    jpus_array = np.empty(num_n_all, dtype=complex)
    yus_array = np.empty(num_n_all, dtype=complex)
    ypus_array = np.empty(num_n_all, dtype=complex)
    kvs_array = np.empty(num_n_all, dtype=complex)
    kpvs_array = np.empty(num_n_all, dtype=complex)
    A1s_array = np.empty(num_n_all, dtype=complex)
    B1s_array = np.empty(num_n_all, dtype=complex)
    A2s_array = np.empty(num_n_all, dtype=complex)
    B2s_array = np.empty(num_n_all, dtype=complex)
    C2s_array = np.empty(num_n_all, dtype=complex)
    D2s_array = np.empty(num_n_all, dtype=complex)
    A3s_array = np.empty(num_n_all, dtype=complex)
    B3s_array = np.empty(num_n_all, dtype=complex)
    Ys_array = np.empty(num_n_all, dtype=complex)
    cdef:
        cdouble[:] xs = xs_array
        cdouble[:] ys = ys_array
        cdouble[:] us = us_array
        cdouble[:] vs = vs_array
        cdouble[:] jxs = jxs_array
        cdouble[:] jpxs = jpxs_array
        cdouble[:] yxs = yxs_array
        cdouble[:] ypxs = ypxs_array
        cdouble[:] iys = iys_array
        cdouble[:] ipys = ipys_array
        cdouble[:] jus = jus_array
        cdouble[:] jpus = jpus_array
        cdouble[:] yus = yus_array
        cdouble[:] ypus = ypus_array
        cdouble[:] kvs = kvs_array
        cdouble[:] kpvs = kpvs_array
        cdouble[:] A1s = A1s_array
        cdouble[:] B1s = B1s_array
        cdouble[:] A2s = A2s_array
        cdouble[:] B2s = B2s_array
        cdouble[:] C2s = C2s_array
        cdouble[:] D2s = D2s_array
        cdouble[:] A3s = A3s_array
        cdouble[:] B3s = B3s_array
        cdouble[:] Ys = Ys_array
    coefs_C(&hs[0], w, &s_all[0], &n_all[0], num_n_all, r, ri, e1, e2,
        &xs[0], &ys[0], &us[0], &vs[0], &jxs[0], &jpxs[0], &yxs[0], &ypxs[0],
        &iys[0], &ipys[0], &jus[0], &jpus[0], &yus[0], &ypus[0], &kvs[0], &kpvs[0],
        &A1s[0], &B1s[0], &A2s[0], &B2s[0], &C2s[0], &D2s[0], &A3s[0], &B3s[0], &Ys[0])
    return (xs_array, ys_array, us_array, vs_array,
            jxs_array, jpxs_array, yxs_array, ypxs_array, iys_array, ipys_array,
            jus_array, jpus_array, yus_array, ypus_array, kvs_array, kpvs_array,
            A1s_array, B1s_array, A2s_array, B2s_array, C2s_array, D2s_array, A3s_array, B3s_array, Ys_array)



# -*- coding: utf-8 -*-
# cython: profile=False
from . cimport (
    INFINITY,
    M_PI,
    cabs,
    carg,
    cdouble,
    cexp,
    cimag,
    creal,
    csqrt,
    log10,
    printf,
    sqrt,
)
from .bessel_utils cimport (
    iv_ivp,
    ive,
    ive_ivpe,
    ive_ivpe_ivppe,
    jv_jvp,
    jve,
    jve_jvpe,
    jve_jvpe_jvppe,
    kv_kvp,
    kve,
    kve_kvpe,
    kve_kvpe_kvppe,
    yv_yvp,
    yve,
    yve_yvpe,
    yve_yvpe_yvppe,
)
from .eig_mat_utils cimport deriv_det2, deriv_det4, det4, solve


cdef cdouble u_func(cdouble h2, cdouble w, cdouble e1, double r) nogil
cdef cdouble v_func(cdouble h2, cdouble w, cdouble e1, double r) nogil
cdef void eig_mat_with_deriv(
    cdouble h2, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, double ri, cdouble[:, ::1] a, cdouble[:, ::1] b
)
cdef void eig_mat(
    cdouble h2, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, double ri, cdouble[:, ::1] a
)
cdef void eig_mat_u_with_deriv(
    cdouble u, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, double ri, cdouble[:, ::1] a, cdouble[:, ::1] b
)
cdef void eig_mat_u(
    cdouble u, cdouble w, str pol, int n, cdouble e1, cdouble e2, double r, double ri, cdouble[:, ::1] a
)
cdef void coefs_C(
    cdouble *hs, cdouble w, long *s_all, long *n_all,
    int num_n_all, double r, double ri, cdouble e1, cdouble e2,
    cdouble *xs, cdouble *ys, cdouble *us, cdouble *vs,
    cdouble *jxs, cdouble *jpxs, cdouble *yxs, cdouble *ypxs, cdouble *iys, cdouble *ipys,
    cdouble *jus, cdouble *jpus, cdouble *yus, cdouble *ypus, cdouble *kvs, cdouble *kpvs,
    cdouble *A1s, cdouble *B1s, cdouble *A2s, cdouble *B2s,
    cdouble *C2s, cdouble *D2s, cdouble *A3s, cdouble *B3s,
    cdouble *Ys) nogil



# -*- coding: utf-8 -*-
# cython: profile=False
from . cimport cabs, cdouble, cimag, creal, csqrt, printf


cdef cdouble det_cython(cdouble[:, ::1] a)
cdef cdouble det4(cdouble[:, ::1] a) nogil
cdef cdouble deriv_det4(cdouble[:, ::1] a, cdouble[:, ::] b) nogil
cdef cdouble deriv_det2(cdouble[:, ::1] a, cdouble[:, ::] b)
cdef void solve(cdouble a[3][3], cdouble b[3], cdouble x[3]) nogil



# -*- coding: utf-8 -*-
# cython: profile=False
from . cimport cdouble
from .bessel_utils cimport (
    jn,
    jn_jnp,
    jn_jnp_jnpp,
    jn_jnp_jnpp_jnppp,
    jnp,
    yn,
    yn_ynp,
    yn_ynp_ynpp,
    yn_ynp_ynpp_ynppp,
    ynp,
)



from __future__ import annotations

from logging import addLevelName

import numpy as np
import ray
import riip
import scipy.special as ssp
from scipy.optimize import minimize, root

from pymwm.cutoff import Cutoff
from pymwm.utils import coax_utils, eig_mat_utils
from pymwm.waveguide import Sampling


class Samples(Sampling):
    """A class defining samples of phase constants of coaxial waveguide
    modes.

    Attributes:
        fill: An instance of Material class for the core
        clad: An instance of Material class for the clad
        size: A float indicating the size of core [um].
        size2: A float indicating the optional size of core [um].
        params: A dict whose keys and values are as follows:
            'wl_max': A float indicating the maximum wavelength [um]
            'wl_min': A float indicating the minimum wavelength [um]
            'wl_imag': A float indicating the minimum value of
                abs(c / f_imag) [um] where f_imag is the imaginary part of
                the frequency.
            'dw': A float indicating frequency interval
                [rad * c / 1um]=[2.99792458e14 rad / s].
            'num_n': An integer indicating the number of orders of modes.
            'num_m': An integer indicating the number of modes in each
                order and polarization.
        ws: A 1D array indicating the real part of the angular frequencies
            to be calculated [rad (c / 1um)]=[2.99792458e14 rad / s].
        wis: A 1D array indicating the imaginary part of the angular
            frequencies to be calculated [rad * (c / 1um)].
        ri: A float indicating inner radius [um].
        r: A float indicating outer radius [um].
        co_list: A list indicating cutoffs for PEC waveguide.
    """

    def __init__(
        self, size: float, fill: dict, clad: dict, params: dict, size2: float
    ) -> None:
        """Init Samples class.

        Args:
            size: A float indicating the outer radius [um]
            fill (dict): Parameters for riip.Material class for the core
            clad (dict): Parameters for riip.Material class for the clad
            params (dict): Keys and values are as follows:
                'wl_max' (float): The maximum wavelength [um].
                    Defaults to 5.0.
                'wl_min' (float): The minimum wavelength [um].
                    Defaults to 0.4.
                'wl_imag' (float): The minimum value of abs(c / f_imag) [um]
                    where f_imag is the imaginary part of the frequency.
                    Defaults to 5.0.
                'dw' (float): The frequency interval [rad c / 1um]=[2.99792458e14 rad / s].
                    Defaults to 1 / 64.
                'num_n' (int): The number of orders of modes.
                'num_m' (int): The number of modes in each order and polarization.
                    (num_m + 1 for TM-like mode, num_m for TE-like mode)
            size2: A float indicating the inner radius [um]
        """
        super().__init__(size, fill, clad, params, size2)
        self.r = size
        self.ri = size2
        self.r_ratio = self.ri / self.r
        num_n = self.params["num_n"]
        num_m = self.params["num_m"]
        co = Cutoff(num_n, num_m)
        self.co_list = []
        for n in range(num_n):
            co_per_n = []
            for pol, m_end in [("M", num_m + 2), ("E", num_m + 1)]:
                for m in range(1, m_end):
                    alpha = (pol, n, m)
                    co_per_n.append(co(alpha, self.r_ratio))
            self.co_list.append(np.array(co_per_n))

    @property
    def shape(self):
        return "coax"

    @property
    def num_all(self):
        return self.params["num_n"] * (2 * self.params["num_m"] + 1)

    def beta2_pec(self, w, n):
        """Return squares of phase constants for a PEC waveguide

        Args:
            w: A complex indicating the angular frequency
            n: A integer indicating the order of the modes.
        Returns:
            h2s: A 1D array indicating squares of phase constants, whose first
                num_mn + 1 elements are for TM-like modes and the rest are for
                TE-like modes.
        """
        w_comp = w.real + 1j * w.imag
        return self.fill(w_comp) * w_comp**2 - self.co_list[n] ** 2 / self.r**2

    def x(
        self, h2: complex | np.ndarray, w: complex, e1: complex
    ) -> complex | np.ndarray:
        val: complex | np.ndarray = (
            (1 + 1j) * np.sqrt(-0.5j * (e1 * w**2 - h2)) * self.ri
        )
        # val: complex | np.ndarray = np.sqrt(e1 * w ** 2 - h2 + 0j) * self.ri
        return val

    def y(
        self, h2: complex | np.ndarray, w: complex, e2: complex
    ) -> complex | np.ndarray:
        val: complex | np.ndarray = (
            (1 - 1j) * np.sqrt(0.5j * (h2 - e2 * w**2)) * self.ri
        )
        # val: complex | np.ndarray = np.sqrt(h2 - e2 * w ** 2 + 0j) * self.ri
        return val

    def u(
        self,
        h2: complex | np.ndarray,
        w: complex | np.ndarray,
        e1: complex | np.ndarray,
    ) -> complex | np.ndarray:
        val: complex | np.ndarray = (
            (1 + 1j) * np.sqrt(-0.5j * (e1 * w**2 - h2)) * self.r
        )
        # val: complex | np.ndarray = np.sqrt(e1 * w ** 2 - h2 + 0j) * self.r
        return val

    def v(
        self,
        h2: complex | np.ndarray,
        w: complex | np.ndarray,
        e2: complex | np.ndarray,
    ) -> complex | np.ndarray:
        val: complex | np.ndarray = (
            (1 - 1j) * np.sqrt(0.5j * (h2 - e2 * w**2)) * self.r
        )
        # val: complex | np.ndarray = np.sqrt(h2 - e2 * w ** 2 + 0j) * self.r
        return val

    def eig_mat(
        self, h2: complex, w: complex, pol: str, n: int, e1: complex, e2: complex
    ) -> tuple[np.ndarray, np.ndarray]:
        """Return matrix of characteristic equation

        Args:
            h2: The square of the phase constant.
            w: The angular frequency
            pol: The polarization
            n: The order of the modes
            e1: The permittivity of the core
            e2: The permittivity of the clad.
        Returns:
            (a, b) (tuple[np.ndarray, np.ndarray]): a: matrix of characteristic equation and b: its defivative.
        """
        h2comp = h2.real + 1j * h2.imag
        w2 = w**2
        hew = h2comp / e2 / w2
        ee = e1 / e2
        x = self.x(h2comp, w, e1)
        y = self.y(h2comp, w, e2)
        u = self.u(h2comp, w, e1)
        v = self.v(h2comp, w, e2)

        if u.imag > 0:
            sign_u = 1
        elif u.imag == 0:
            sign_u = 0
        else:
            sign_u = -1

        if v.real > 0:
            sign_v = 1
        elif v.real == 0:
            sign_v = 0
        else:
            sign_v = -1
        ph_u = np.exp(1j * sign_u * u.real)

        ju = ssp.jve(n, u) * ph_u
        jpu = -ssp.jve(n + 1, u) * ph_u + n / u * ju
        jppu = -jpu / u - (1 - n**2 / u**2) * ju

        yu = ssp.yve(n, u) * ph_u
        ypu = -ssp.yve(n + 1, u) * ph_u + n / u * yu
        yppu = -ypu / u - (1 - n**2 / u**2) * yu

        kv = ssp.kve(n, v)
        kpv = -ssp.kve(n + 1, v) + n / v * kv
        kppv = -kpv / v + (1 + n**2 / v**2) * kv

        ph_x = np.exp(1j * sign_u * x.real)
        jx = ssp.jve(n, x) * ph_x
        jpx = -ssp.jve(n + 1, x) * ph_x + n / x * jx
        jppx = -jpx / x - (1 - n**2 / x**2) * jx

        yx = ssp.yve(n, x) * ph_x
        ypx = -ssp.yve(n + 1, x) * ph_x + n / x * yx
        yppx = -ypx / x - (1 - n**2 / x**2) * yx

        ph_y = np.exp(-1j * sign_v * y.imag)
        iy = ssp.ive(n, y) * ph_y
        ipy = ssp.ive(n + 1, y) * ph_y + n / y * iy
        ippy = -ipy / y + (1 + n**2 / y**2) * iy

        du_dh2 = -self.r**2 / (2 * u)
        dv_dh2 = self.r**2 / (2 * v)
        dx_dh2 = -self.ri**2 / (2 * x)
        dy_dh2 = self.ri**2 / (2 * y)

        nuv = n * (v / u + u / v)
        dnuv_du = n * (-v / u**2 + 1 / v)
        dnuv_dv = n * (-u / v**2 + 1 / u)

        nxy = n * (y / x + x / y)
        dnxy_dx = n * (-y / x**2 + 1 / y)
        dnxy_dy = n * (-x / y**2 + 1 / x)

        a = np.array(
            [
                [
                    jpu * kv * v + kpv * ju * u,
                    ypu * kv * v + kpv * yu * u,
                    nuv * ju * kv,
                    nuv * yu * kv,
                ],
                [
                    jpx * y + ipy / iy * jx * x,
                    ypx * y + ipy / iy * yx * x,
                    nxy * jx,
                    nxy * yx,
                ],
                [
                    hew * nuv * ju * kv,
                    hew * nuv * yu * kv,
                    ee * jpu * kv * v + kpv * ju * u,
                    ee * ypu * kv * v + kpv * yu * u,
                ],
                [
                    hew * nxy * jx,
                    hew * nxy * yx,
                    ee * jpx * y + ipy / iy * jx * x,
                    ee * ypx * y + ipy / iy * yx * x,
                ],
            ]
        )

        da_du = np.array(
            [
                [
                    jppu * kv * v + kpv * (jpu * u + ju) + 1j * sign_u * a[0, 0],
                    yppu * kv * v + kpv * (ypu * u + yu) + 1j * sign_u * a[0, 1],
                    dnuv_du * ju * kv + nuv * jpu * kv + 1j * sign_u * a[0, 2],
                    dnuv_du * yu * kv + nuv * ypu * kv + 1j * sign_u * a[0, 3],
                ],
                [0, 0, 0, 0],
                [
                    hew * (dnuv_du * ju + nuv * jpu) * kv + 1j * sign_u * a[2, 0],
                    hew * (dnuv_du * yu + nuv * ypu) * kv + 1j * sign_u * a[2, 1],
                    ee * jppu * kv * v + kpv * (jpu * u + ju) + 1j * sign_u * a[2, 2],
                    ee * yppu * kv * v + kpv * (ypu * u + yu) + 1j * sign_u * a[2, 3],
                ],
                [0, 0, 0, 0],
            ]
        )

        da_dv = np.array(
            [
                [
                    jpu * (kpv * v + kv) + kppv * ju * u + a[0, 0],
                    ypu * (kpv * v + kv) + kppv * yu * u + a[0, 1],
                    (dnuv_dv * kv + nuv * kpv) * ju + a[0, 2],
                    (dnuv_dv * kv + nuv * kpv) * yu + a[0, 3],
                ],
                [0, 0, 0, 0],
                [
                    hew * (dnuv_dv * kv + nuv * kpv) * ju + a[2, 0],
                    hew * (dnuv_dv * kv + nuv * kpv) * yu + a[2, 1],
                    ee * jpu * (kpv * v + kv) + kppv * ju * u + a[2, 2],
                    ee * ypu * (kpv * v + kv) + kppv * yu * u + a[2, 3],
                ],
                [0, 0, 0, 0],
            ]
        )

        da_dx = np.array(
            [
                [0, 0, 0, 0],
                [
                    jppx * y + ipy / iy * (jpx * x + jx) + 1j * sign_u * a[1, 0],
                    yppx * y + ipy / iy * (ypx * x + yx) + 1j * sign_u * a[1, 1],
                    dnxy_dx * jx + nxy * jpx + 1j * sign_u * a[1, 2],
                    dnxy_dx * yx + nxy * ypx + 1j * sign_u * a[1, 3],
                ],
                [0, 0, 0, 0],
                [
                    hew * (dnxy_dx * jx + nxy * jpx) + 1j * sign_u * a[3, 0],
                    hew * (dnxy_dx * yx + nxy * ypx) + 1j * sign_u * a[3, 1],
                    ee * jppx * y + ipy / iy * (jpx * x + jx) + 1j * sign_u * a[3, 2],
                    ee * yppx * y + ipy / iy * (ypx * x + yx) + 1j * sign_u * a[3, 3],
                ],
            ]
        )

        da_dy = np.array(
            [
                [0, 0, 0, 0],
                [
                    jpx + (ippy / iy - ipy**2 / iy**2) * jx * x,
                    ypx + (ippy / iy - ipy**2 / iy**2) * yx * x,
                    dnxy_dy * jx,
                    dnxy_dy * yx,
                ],
                [0, 0, 0, 0],
                [
                    hew * dnxy_dy * jx,
                    hew * dnxy_dy * yx,
                    ee * jpx + (ippy / iy - ipy**2 / iy**2) * jx * x,
                    ee * ypx + (ippy / iy - ipy**2 / iy**2) * yx * x,
                ],
            ]
        )

        b = (
            da_du * du_dh2
            + da_dv * dv_dh2
            + da_dx * dx_dh2
            + da_dy * dy_dh2
            + np.array(
                [
                    [0, 0, 0, 0],
                    [0, 0, 0, 0],
                    [
                        1.0 / (e2 * w2) * nuv * ju * kv,
                        1.0 / (e2 * w2) * nuv * yu * kv,
                        0,
                        0,
                    ],
                    [
                        1.0 / (e2 * w2) * nxy * jx,
                        1.0 / (e2 * w2) * nxy * yx,
                        0,
                        0,
                    ],
                ]
            )
        )
        return a, b

    def eig_eq(
        self,
        h2vec: np.ndarray,
        w: complex,
        pol: str,
        n: int,
        e1: complex,
        e2: complex,
        roots: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Return the value of the characteristic equation

        Args:
            h2vec: The real and imaginary parts of the square of propagation constant.
            w: The angular frequency
            pol: The polarization
            n: The order of the modes
            e1: The permittivity of the core
            e2: The permittivity of the clad.
        Returns:
            val: A complex indicating the left-hand value of the characteristic
                equation.
        """
        num = len(roots)
        h2 = h2vec[0] + h2vec[1] * 1j
        a, b = self.eig_mat(h2, w, pol, n, e1, e2)
        if n == 0:
            if pol == "E":
                f = a[0, 0] * a[1, 1] - a[0, 1] * a[1, 0]
                fp = eig_mat_utils.deriv_det2_cython(
                    np.ascontiguousarray(a[:2, :2]), np.ascontiguousarray(b[:2, :2])
                )
            else:
                f = a[2, 2] * a[3, 3] - a[2, 3] * a[3, 2]
                fp = eig_mat_utils.deriv_det2_cython(
                    np.ascontiguousarray(a[2:, 2:]), np.ascontiguousarray(b[2:, 2:])
                )
        else:
            f = np.linalg.det(a)
            fp = eig_mat_utils.deriv_det4_cython(a, b)
        denom = 1.0
        dd = 0.0
        for i in range(num):
            denom *= (h2 - roots[i]) / roots[i]
            ddi = -roots[i] / (h2 - roots[i]) ** 2
            for j in range(num):
                if j != i:
                    ddi /= (h2 - roots[j]) / roots[j]
            dd += ddi
        fp = fp / denom + f * dd
        f /= denom
        return np.array([f.real, f.imag]), np.array(
            [[fp.real, fp.imag], [-fp.imag, fp.real]]
        )

    def beta2(
        self, w: complex, n: int, e1: complex, e2: complex, xis: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        """Return roots and convergences of the characteristic equation

        Args:
            w (complex): Angular frequency.
            n (int): Order of the mode
            e1 (complex): Permittivity of tha core.
            e2 (complex): Permittivity of tha clad.
            xis (np.ndarray): Initial approximations for the roots
                whose number of elements is 2*num_m+1.
        Returns:
            xs: A 1D array indicating the roots, whose length is 2*num_m+1.
            success: A 1D array indicating the convergence information for xs.
        """
        if self.clad.label == "PEC":
            xs = self.beta2_pec(w, n)
            return xs, np.ones_like(xs, dtype=bool)
        num_m = self.params["num_m"]
        roots: list[complex] = []
        vals = []
        successes: list[bool] = []

        for i, xi in enumerate(xis):
            if i < num_m + 1:
                pol = "M"
            else:
                pol = "E"
                if n == 0 and i == num_m + 1:
                    roots = []
            args = (w, pol, n, e1, e2, self.r, self.ri, np.array(roots, dtype=complex))
            result = root(
                coax_utils.eig_eq_with_jac,
                np.array([xi.real, xi.imag]),
                args=args,
                jac=True,
                method="hybr",
                options={"col_deriv": True},
            )
            x = result.x[0] + result.x[1] * 1j
            success = result.success
            if not success and np.hypot(*result.fun) < 1e-8:
                success = True
            if success:
                roots.append(x)
            successes.append(success)
            vals.append(x)
        return np.array(vals), np.array(successes)

    def beta2_with_u(
        self, w: complex, n: int, e1: complex, e2: complex, xis: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        """Return roots and convergences of the characteristic equation

        Args:
            w (complex): Angular frequency.
            n (int): Order of the mode
            e1 (complex): Permittivity of tha core.
            e2 (complex): Permittivity of tha clad.
            xis (np.ndarray): Initial approximations for the roots
                whose number of elements is 2*num_m+1.
        Returns:
            xs: A 1D array indicating the roots, whose length is 2*num_m+1.
            success: A 1D array indicating the convergence information for xs.
        """
        if self.clad.label == "PEC":
            xs = self.beta2_pec(w, n)
            return xs, np.ones_like(xs, dtype=bool)
        num_m = self.params["num_m"]
        roots: list[complex] = []
        vals = []
        success: list[bool] = []

        for i, xi in enumerate(xis):
            u0 = self.u(xi, w, e1)
            if i < num_m + 1:
                pol = "M"
            else:
                pol = "E"
                if n == 0 and i == num_m + 1:
                    roots = []
            args = (w, pol, n, e1, e2, self.r, self.ri, np.array(roots, dtype=complex))
            result = root(
                coax_utils.eig_eq_u_with_jac,
                np.array([u0.real, u0.imag]),
                args=args,
                jac=True,
                method="hybr",
                options={"col_deriv": True},
            )
            u = result.x[0] + 1j * result.x[1]
            if np.hypot(*result.fun) < 1e-8:
                roots.append(u)
                success.append(True)
            else:
                # print(_w, result.success, result.fun)
                success.append(False)
            vals.append(e1 * w**2 - u**2 / self.r**2)
        return np.array(vals), np.array(success)

    @staticmethod
    def beta_from_beta2(x):
        return (1 + 1j) * np.sqrt(-0.5j * x)
        # return np.sqrt(x)
        # return 1j * np.sqrt(-x)

    def beta2_w_min(self, n):
        """Return roots and convergences of the characteristic equation at
            the lowest angular frequency, ws[0].

        Args:
            n: A integer indicating the order of the mode
        Returns:
            xs: A 1D array indicating the roots, whose length is 2*num_m+1.
            success: A 1D array indicating the convergence information for xs.
        """
        if self.clad.label == "PEC":
            xs = self.beta2_pec(self.ws[0], n)
            success = np.ones_like(xs, dtype=bool)
            return xs, success
        w_0 = 2 * np.pi / 10.0
        e1 = self.fill(w_0)
        e2_0 = self.clad(w_0) * 1000
        de2 = (self.clad(w_0) - e2_0) / 5000
        xs, success = self.beta2(w_0, n, e1, e2_0, self.beta2_pec(w_0, n))
        xs0 = xs1 = xs
        success = np.ones_like(xs0, dtype=bool)
        for i in range(5001):
            e2 = e2_0 + de2 * i
            xis = 2 * xs1 - xs0
            xs, success = self.beta2(w_0, n, e1, e2, xis)
            xs = np.where(success, xs, xs1)
            xs0 = xs1
            xs1 = xs
        xs0 = xs1
        dw = (self.ws[0] - w_0) / 500
        for i in range(501):
            w = w_0 + dw * i
            e1 = self.fill(w)
            e2 = self.clad(w)
            xis = 2 * xs1 - xs0
            xs, success = self.beta2(w, n, e1, e2, xis)
            xs = np.where(success, xs, xs1)
            xs0 = xs1
            xs1 = xs
        return xs, success

    def betas_convs(self, xs_success_list: list) -> tuple[dict, dict]:
        num_n = self.params["num_n"]
        num_m = self.params["num_m"]
        betas = {}
        convs = {}
        for n in range(num_n):
            xs_array, success_array = xs_success_list[n]
            for m in range(1, num_m + 2):
                betas[("M", n, m)] = np.zeros(
                    (len(self.ws), len(self.wis)), dtype=complex
                )
                convs[("M", n, m)] = np.zeros((len(self.ws), len(self.wis)), dtype=bool)
            for m in range(1, num_m + 1):
                betas[("E", n, m)] = np.zeros(
                    (len(self.ws), len(self.wis)), dtype=complex
                )
                convs[("E", n, m)] = np.zeros((len(self.ws), len(self.wis)), dtype=bool)
            for iwi in range(len(self.wis)):
                for iwr in range(len(self.ws)):
                    for i in range(num_m + 1):
                        x = xs_array[iwr, iwi][i]
                        betas[("M", n, i + 1)][iwr, iwi] = self.beta_from_beta2(x)
                        convs[("M", n, i + 1)][iwr, iwi] = success_array[iwr, iwi][i]
                    for i in range(num_m):
                        x = xs_array[iwr, iwi][i + num_m + 1]
                        betas[("E", n, i + 1)][iwr, iwi] = self.beta_from_beta2(x)
                        convs[("E", n, i + 1)][iwr, iwi] = success_array[iwr, iwi][
                            i + num_m + 1
                        ]
        return betas, convs

    def beta2_adaptive(
        self,
        real_or_imag: str,
        w: complex,
        n: int,
        w0: complex,
        x0s: np.ndarray,
        dx0=0.01,
        dw_min=2 ** (-24),
    ) -> tuple[np.ndarray, np.ndarray]:
        """Return roots and convergences of the characteristic equation

        Args:
            real_or_imag (str): "real" or "imag".
            w0 (complex): Angular frequency of previous step.
            w (complex): Angular frequency.
            n (int): Order of the mode
            x0s (np.ndarray): Initial approximations for the roots
                whose number of elements is 2*num_m+1.
        Returns:
            xs: A 1D array indicating the roots, whose length is 2*num_m+1.
            success: A 1D array indicating the convergence information for xs.
        """
        if self.clad.label == "PEC":
            xs = self.beta2_pec(w, n)
            return xs, np.ones_like(xs, dtype=bool)

        def func(_w, _x0s) -> tuple[np.ndarray, np.ndarray]:
            num_m = self.params["num_m"]
            roots: list[complex] = []
            vals = []
            successes: list[bool] = []
            for i, _x0 in enumerate(_x0s):
                if i < num_m + 1:
                    pol = "M"
                else:
                    pol = "E"
                    if n == 0 and i == num_m + 1:
                        roots = []
                args = (
                    _w,
                    pol,
                    n,
                    self.fill(_w),
                    self.clad(_w),
                    self.r,
                    self.ri,
                    np.array(roots, dtype=complex),
                )
                result = root(
                    coax_utils.eig_eq_with_jac,
                    np.array([_x0.real, _x0.imag]),
                    args=args,
                    jac=True,
                    method="hybr",
                    options={"col_deriv": True},
                )
                x = result.x[0] + 1j * result.x[1]
                success = result.success
                if not success and np.hypot(*result.fun) < 1e-8:
                    success = True
                if success:
                    roots.append(x)
                vals.append(x)
                successes.append(success)
            return np.array(vals), np.array(successes)

        _w0, _x0s = w0, x0s
        _w1 = w
        _x1s, successes = func(_w1, x0s)
        while True:
            if not np.all(successes):
                factor = 0
            else:
                factor = min(dx0 / np.abs(_x1s - _x0s).max(), 2)
            dw = max(abs(_w1 - _w0) * factor, dw_min)
            if real_or_imag == "real":
                _w = _w0 + dw
            else:
                _w = _w0 - 1j * dw
            _xis = _x0s + (_x1s - _x0s) * factor
            xs, successes = func(_w, _xis)
            _w0, _x0s = _w1, _x1s
            _w1, _x1s = _w, xs
            if real_or_imag == "real":
                if _w1.real > w.real:
                    break
            else:
                if _w1.imag < w.imag:
                    break
        return func(w, _x1s)

    def __call__(self, n: int):
        """Return a dict of the roots of the characteristic equation

        Args:
            n: A integer indicating the order of the mode
        Returns:
            betas: A dict containing arrays of roots, whose key is as follows:
                (pol, n, m):
                    pol: 'E' or 'M' indicating the polarization.
                    n: A integer indicating the order of the mode.
                    m: A integer indicating the ordinal of the mode in the same
                        order.
            convs: A dict containing the convergence information for betas,
                whose key is the same as above.
        """
        num_m = self.params["num_m"]
        xs_array = np.zeros((len(self.ws), len(self.wis), 2 * num_m + 1), dtype=complex)
        success_array = np.zeros(
            (len(self.ws), len(self.wis), 2 * num_m + 1), dtype=bool
        )
        iwr = iwi = 0
        wi = self.wis[iwi]
        xis, success = self.beta2_w_min(n)
        xs_array[iwr, iwi] = xis
        success_array[iwr, iwi] = success
        xs0 = xs1 = xis
        for iwr in range(1, len(self.ws)):
            wr = self.ws[iwr]
            w = wr + 1j * wi
            e1 = self.fill(w)
            e2 = self.clad(w)
            xis = 2 * xs1 - xs0
            xs, success = self.beta2(w, n, e1, e2, xis)
            xs = np.where(success, xs, xis)
            xs_array[iwr, iwi] = xs
            success_array[iwr, iwi] = success
            xs0 = xs1
            xs1 = xs
        for iwr in range(len(self.ws)):
            wr = self.ws[iwr]
            xs0 = xs1 = xis = xs_array[iwr, 0]
            for iwi in range(1, len(self.wis)):
                wi = self.wis[iwi]
                w = wr + 1j * wi
                e1 = self.fill(w)
                e2 = self.clad(w)
                xis = 2 * xs1 - xs0
                xs, success = self.beta2(w, n, e1, e2, xis)
                xs = np.where(success, xs, xis)
                xs_array[iwr, iwi] = xs
                success_array[iwr, iwi] = success
                xs0 = xs1
                xs1 = xs
        return xs_array, success_array

    def wr_sampling(self, n: int) -> tuple[np.ndarray, np.ndarray]:
        num_m = self.params["num_m"]
        xs_array = np.zeros((len(self.ws), 2 * num_m + 1), dtype=complex)
        success_array = np.zeros((len(self.ws), 2 * num_m + 1), dtype=bool)
        iwr = 0
        x0s, success = self.beta2_w_min(n)
        xs_array[iwr] = x0s
        success_array[iwr] = success
        for iwr in range(1, len(self.ws)):
            w = self.ws[iwr]
            # print(iwr, w)
            w0 = self.ws[iwr - 1]
            xs, success = self.beta2_adaptive("real", w, n, w0, x0s)
            xs_array[iwr] = xs
            success_array[iwr] = success
            x0s = xs
        return xs_array, success_array

    def wi_sampling(
        self, args: tuple[int, int, np.ndarray, np.ndarray]
    ) -> tuple[np.ndarray, np.ndarray]:
        n, iwr, x0s, success = args
        num_m = self.params["num_m"]
        xs_array = np.zeros((len(self.wis), 2 * num_m + 1), dtype=complex)
        success_array = np.zeros((len(self.wis), 2 * num_m + 1), dtype=bool)
        wr = self.ws[iwr]
        iwi = 0
        xs_array[iwi] = x0s
        success_array[iwi] = success
        for iwi in range(1, len(self.wis)):
            wi = self.wis[iwi]
            # print(iwi, wi)
            wi0 = self.wis[iwi - 1]
            xs, success = self.beta2_adaptive(
                "imag", wr + 1j * wi, n, wr + 1j * wi0, x0s
            )
            xs_array[iwi] = xs
            success_array[iwi] = success
            x0s = xs
        return xs_array, success_array


class SamplesLowLoss(Samples):
    """A class defining samples of phase constants of coaxial waveguide
    modes in a virtual low-loss clad waveguide by subclassing the Samples class.

    Attributes:
        fill: An instance of Material class for the core
        clad: An instance of Material class for the clad
        size: A float indicating the size of core [um].
        size2: A float indicating the optional size of core [um].
        params: A dict whose keys and values are as follows:
            'wl_max': A float indicating the maximum wavelength [um]
            'wl_min': A float indicating the minimum wavelength [um]
            'wl_imag': A float indicating the minimum value of
                abs(c / f_imag) [um] where f_imag is the imaginary part of
                the frequency.
            'dw': A float indicating frequency interval
                [rad * c / 1um]=[2.99792458e14 rad / s].
            'num_n': An integer indicating the number of orders of modes.
            'num_m': An integer indicating the number of modes in each
                order and polarization.
        ws: A 1D array indicating the real part of the angular frequencies
            to be calculated [rad (c / 1um)]=[2.99792458e14 rad / s].
        wis: A 1D array indicating the imaginary part of the angular
            frequencies to be calculated [rad * (c / 1um)].
        ri: A float indicating inner radius [um].
        r: A float indicating outer radius [um].
        co_list: A list indicating cutoffs for PEC waveguide.
    """

    def __init__(self, size: float, fill: dict, clad: dict, params: dict, size2):
        """Init Samples class.

        size (float): The radius of the cross section [um]
        fill (dict): Parameters for riip.Material class for the core
        clad (dict): Parameters for riip.Material class for the clad
        params (dict): Keys and values are as follows:
            'wl_max' (float): The maximum wavelength [um].
                Defaults to 5.0.
            'wl_min' (float): The minimum wavelength [um].
                Defaults to 0.4.
            'wl_imag' (float): The minimum value of abs(c / f_imag) [um]
                where f_imag is the imaginary part of the frequency.
                Defaults to 5.0.
            'dw' (float): The frequency interval [rad c / 1um]=[2.99792458e14 rad / s].
                Defaults to 1 / 64.
            'num_n' (int): The number of orders of modes.
            'num_m' (int): The number of modes in each order and polarization.
        """
        super().__init__(size, fill, clad, params, size2)

    def betas_convs(self, xs_success_list):
        num_iwr = len(self.ws)
        num_iwi = len(self.wis)
        num_n = self.params["num_n"]
        num_m = self.params["num_m"]
        betas = {}
        convs = {}
        for n in range(num_n):
            for m in range(1, num_m + 2):
                betas[("M", n, m)] = np.zeros(
                    (len(self.ws), len(self.wis)), dtype=complex
                )
                convs[("M", n, m)] = np.zeros((len(self.ws), len(self.wis)), dtype=bool)
            for m in range(1, num_m + 1):
                betas[("E", n, m)] = np.zeros(
                    (len(self.ws), len(self.wis)), dtype=complex
                )
                convs[("E", n, m)] = np.zeros((len(self.ws), len(self.wis)), dtype=bool)
        for iwr in range(num_iwr):
            for iwi in range(num_iwi):
                j = iwr * num_iwi + iwi
                for n in range(num_n):
                    for i in range(num_m + 1):
                        x = xs_success_list[j][0][n][i]
                        betas[("M", n, i + 1)][iwr, iwi] = self.beta_from_beta2(x)
                        convs[("M", n, i + 1)][iwr, iwi] = xs_success_list[j][1][n][i]
                    for i in range(num_m):
                        x = xs_success_list[j][0][n][i + num_m + 1]
                        betas[("E", n, i + 1)][iwr, iwi] = self.beta_from_beta2(x)
                        convs[("E", n, i + 1)][iwr, iwi] = xs_success_list[j][1][n][
                            i + num_m + 1
                        ]
        return betas, convs


@ray.remote
class SamplesForRay(Samples):
    """A derived class in order to create ray actor."""

    def __init__(self, size: float, fill: dict, clad: dict, params: dict, size2: float):
        super().__init__(size, fill, clad, params, size2)


@ray.remote
class SamplesLowLossForRay(SamplesLowLoss):
    """A derived class in order to create ray actor."""

    def __init__(self, size: float, fill: dict, clad: dict, params: dict, size2: float):
        super().__init__(size, fill, clad, params, size2)

    def task(self, arg: tuple[int, int, list[np.ndarray]]):
        """Return a dict of the roots of the characteristic equation

        Args:
            arg: (iwr, iwi, xis_list)
                iwr: The ordinal of the Re(w).
                iwi: The ordinal of the Im(w).
                xis_list: The initial guess of roots whose length is 2*num_m+1
        Returns:
            xs_list: A list of num_n 1D arrays indicating the roots, whose
                length is 2*num_m+1
            success_list: A list of num_n 1D arrays indicating the convergence
                information for xs, whose length is 2*num_m+1
        """
        iwr, iwi, xis_list = arg
        im_factor = self.clad.im_factor
        wr = self.ws[iwr]
        wi = self.wis[iwi]
        w = wr + 1j * wi
        e1 = self.fill(w)
        xs_list = []
        success_list = []
        for n, x0s in enumerate(xis_list):
            xis = xs = x0s
            success = np.ones_like(xs, dtype=bool)
            for i in range(1, 8):
                self.clad.im_factor = 0.5**i
                if i == 7 or self.clad.im_factor < im_factor:
                    self.clad.im_factor = im_factor
                e2 = self.clad(w)
                xs, success = self.beta2(w, n, e1, e2, xis)
                for _, ok in enumerate(success):
                    if not ok:
                        xs[_] = xis[_]
                xis = xs
                if self.clad.im_factor == im_factor:
                    break
            xs_list.append(xs)
            success_list.append(success)
        return xs_list, success_list



# -*- coding: utf-8 -*-
# cython: profile=False
cimport cython


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double jn(int n, double z) nogil:
    return jv(n, z)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double jnp(int n, double z) nogil:
    return 0.5 * (jv(n - 1, z) - jv(n + 1, z))

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void jn_jnp(int n, double z, double vals[2]) nogil:
    cdef:
         double j = jn(n, z)
    vals[0] = j
    vals[1] = -jn(n + 1, z) + n * j / z

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void jn_jnp_jnpp(int n, double z, double vals[3]) nogil:
    cdef:
         double j = jn(n, z)
         double jp = -jn(n + 1, z) + n * j / z
    vals[0] = j
    vals[1] = jp
    vals[2] = -jp / z - (1 - n ** 2 / z ** 2) * j

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void jn_jnp_jnpp_jnppp(int n, double z, double vals[4]) nogil:
    cdef:
         double j = jn(n, z)
         double jp = -jn(n + 1, z) + n * j / z
         double jpp = -jp / z - (1 - n ** 2 / z ** 2) * j
    vals[0] = j
    vals[1] = jp
    vals[2] = jpp
    vals[3] = -jpp / z - jp + (n ** 2 + 1) * jp / z ** 2 - 2 * n ** 2 / z ** 3 * j


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double yn(int n, double z) nogil:
    return yv(n, z)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double ynp(int n, double z) nogil:
    return 0.5 * (yn(n - 1, z) - yn(n + 1, z))

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void yn_ynp(int n, double z, double vals[2]) nogil:
    cdef:
         double y = yn(n, z)
    vals[0] = y
    vals[1] = -yn(n + 1, z) + n * y / z

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void yn_ynp_ynpp(int n, double z, double vals[3]) nogil:
    cdef:
         double y = yn(n, z)
         double yp = -yn(n + 1, z) + n * y / z
    vals[0] = y
    vals[1] = yp
    vals[2] = -yp / z - (1 - n ** 2 / z ** 2) * y

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void yn_ynp_ynpp_ynppp(int n, double z, double vals[4]) nogil:
    cdef:
         double y = yn(n, z)
         double yp = -yn(n + 1, z) + n * y / z
         double ypp = -yp / z - (1 - n ** 2 / z ** 2) * y
    vals[0] = y
    vals[1] = yp
    vals[2] = ypp
    vals[3] = -ypp / z - yp + (n ** 2 + 1) * yp / z ** 2 - 2 * n ** 2 / z ** 3 * y


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void jv_jvp(int n, cdouble z, cdouble vals[2]) nogil:
    cdef:
         cdouble j = jv(n, z)
    vals[0] = j
    vals[1] = -jv(n + 1, z) + n * j / z


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void yv_yvp(int n, cdouble z, cdouble vals[2]) nogil:
    cdef:
         cdouble y = yv(n, z)
    vals[0] = y
    vals[1] = -yv(n + 1, z) + n * y / z


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def yv_yvp_cython(int n, cdouble z):
    cdef cdouble vals[2]
    yv_yvp(n, z, vals)
    return vals[0], vals[1]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void kv_kvp(int n, cdouble z, cdouble vals[2]) nogil:
    cdef:
        cdouble k = kv(n, z)
    vals[0] = k
    vals[1] = -kv(n + 1, z) + n * k / z


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void iv_ivp(int n, cdouble z, cdouble vals[2]) nogil:
    cdef:
        cdouble i = iv(n, z)
    vals[0] = i
    vals[1] = iv(n + 1, z) + n * i / z


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void jve_jvpe_jvppe(int n, cdouble z, cdouble vals[3]) nogil:
    """ jv * exp(1j * z), jvp * exp(1j * z), jvpp * exp(1j * z) """
    cdef int sign
    if cimag(z) > 0:
        sign = 1
    elif cimag(z) == 0 :
        sign = 0
    else:
        sign = -1
    cdef:
        cdouble ph = cexp(1j * sign * creal(z))
        cdouble j = jve(n, z) * ph
    vals[0] = j
    vals[1] = (jve(n - 1, z) - jve(n + 1, z)) * ph / 2
    vals[2] = (jve(n - 2, z) + jve(n + 2, z)) * ph / 4 - j / 2


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void yve_yvpe_yvppe(int n, cdouble z, cdouble vals[3]) nogil:
    """ yv * exp(1j * z), yvp * exp(1j * z), yvpp * exp(1j * z) """
    cdef int sign
    if cimag(z) > 0:
        sign = 1
    elif cimag(z) == 0 :
        sign = 0
    else:
        sign = -1
    cdef:
        cdouble ph = cexp(1j * sign * creal(z))
        cdouble y = yve(n, z) * ph
    vals[0] = y
    vals[1] = (yve(n - 1, z) - yve(n + 1, z)) * ph / 2
    vals[2] = (yve(n - 2, z) + yve(n + 2, z)) * ph / 4 - y / 2


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void ive_ivpe_ivppe(int n, cdouble z, cdouble vals[3]) nogil:
    """ iv * exp(-z), ivp * exp(-z), ivpp * exp(-z) """
    cdef int sign
    if creal(z) > 0:
        sign = 1
    elif creal(z) == 0 :
        sign = 0
    else:
        sign = -1
    cdef:
        cdouble ph = cexp(-1j * sign * cimag(z))
        cdouble i = ive(n, z) * ph
    vals[0] = i
    vals[1] = (ive(n - 1, z) + ive(n + 1, z)) * ph / 2
    vals[2] = (ive(n - 2, z) + ive(n + 2, z)) * ph / 4 + i / 2


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void kve_kvpe_kvppe(int n, cdouble z, cdouble vals[3]) nogil:
    """ kv * exp(z), kvp * exp(z), kvpp * exp(z) """
    cdef:
        cdouble k = kve(n, z)
    vals[0] = k
    vals[1] = - (kve(n - 1, z) + kve(n + 1, z)) / 2
    vals[2] = (kve(n - 2, z) + kve(n + 2, z)) / 4 + k / 2


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void jve_jvpe(int n, cdouble z, cdouble vals[2]) nogil:
    """ jv * exp(1j * z), jvp * exp(1j * z) """
    cdef int sign
    if cimag(z) > 0:
        sign = 1
    elif cimag(z) == 0 :
        sign = 0
    else:
        sign = -1
    cdef:
        cdouble ph = cexp(1j * sign * creal(z))
    vals[0] = jve(n, z) * ph
    vals[1] = (jve(n - 1, z) - jve(n + 1, z)) * ph / 2


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void yve_yvpe(int n, cdouble z, cdouble vals[2]) nogil:
    """ yv * exp(1j * z), yvp * exp(1j * z) """
    cdef int sign
    if cimag(z) > 0:
        sign = 1
    elif cimag(z) == 0 :
        sign = 0
    else:
        sign = -1
    cdef:
        cdouble ph = cexp(1j * sign * creal(z))
    vals[0] = yve(n, z) * ph
    vals[1] = (yve(n - 1, z) - yve(n + 1, z)) * ph / 2


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void ive_ivpe(int n, cdouble z, cdouble vals[2]) nogil:
    """ iv * exp(-z), ivp * exp(-z) """
    cdef int sign
    if creal(z) > 0:
        sign = 1
    elif creal(z) == 0 :
        sign = 0
    else:
        sign = -1
    cdef:
        cdouble ph = cexp(-1j * sign * cimag(z))
    vals[0] = ive(n, z) * ph
    vals[1] = (ive(n - 1, z) + ive(n + 1, z)) * ph / 2


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void kve_kvpe(int n, cdouble z, cdouble vals[2]) nogil:
    """ kv * exp(z), kvp * exp(z) """
    vals[0] = kve(n, z)
    vals[1] = - (kve(n - 1, z) + kve(n + 1, z)) / 2



from __future__ import annotations

import cmath
import enum
from logging import getLogger

import numpy as np
import psutil
import ray

from pymwm.utils import slit_utils
from pymwm.waveguide import Database, Sampling, Waveguide

from .samples import Samples, SamplesForRay, SamplesLowLoss, SamplesLowLossForRay

logger = getLogger(__package__)


class Slit(Waveguide):
    """A class defining a slit waveguide."""

    def __init__(self, params):
        """Init Slit class.

        Args:
            params: A dict whose keys and values are as follows:
                'core': A dict of the setting parameters of the core:
                    'shape': A string indicating the shape of the core.
                    'size': A float indicating the width of the slit [um].
                    'fill': A dict of the parameters of the core Material.
                'clad': A dict of the parameters of the clad Material.
                'bounds': A dict indicating the bounds of database.interpolation
                    and its keys and values are as follows:
                    'wl_max': A float indicating the maximum wavelength [um]
                    'wl_min': A float indicating the minimum wavelength [um]
                    'wl_imag': A float indicating the maximum value of
                        abs(c / f_imag) [um] where f_imag is the imaginary part
                        of the frequency.
                'modes': A dict of the settings for calculating modes:
                    'wl_max': A float indicating the maximum wavelength [um]
                        (default: 5.0)
                    'wl_min': A float indicating the minimum wavelength [um]
                        (default: 0.4)
                    'wl_imag': A float indicating the maximum value of
                        abs(c / f_imag) [um] where f_imag is the imaginary part
                        of the frequency. (default: 5.0)
                    'dw': A float indicating frequency interval
                        [rad c / 1um]=[2.99792458e14 rad / s]
                        (default: 1 / 64).
                    'num_n': An integer indicating the number of orders of
                        modes.
                    'ls': A list of characters chosen from "h" (horizontal
                        polarization) and "v" (vertical polarization).
                        In the slit case, "h" ("v") corresponds to TE (TM)
                        polarization.
        """
        num_m = params["modes"].setdefault("num_m", 1)
        if num_m != 1:
            logger.warning(
                "num_m must be 1 if shape is slit." + "The set value is ignored."
            )
            params["modes"]["num_m"] = 1
        super().__init__(params)

    def get_alphas(self, alpha_list: list[tuple[str, int, int]]) -> dict:
        alphas: dict = {"h": [], "v": []}
        for alpha in [("E", n, 1) for n in range(1, self.num_n)]:
            if alpha in alpha_list:
                alphas["v"].append(alpha)
        for alpha in [("M", n, 1) for n in range(self.num_n)]:
            if alpha in alpha_list:
                alphas["h"].append(alpha)
        return alphas

    def merge_even_and_odd_data(
        self, even: tuple[np.ndarray, np.ndarray], odd: tuple[np.ndarray, np.ndarray]
    ) -> tuple[np.ndarray, np.ndarray]:
        xs_e, success_e = even
        xs_o, success_o = odd
        num_n_e = xs_e.shape[2]
        num_n_o = xs_o.shape[2]
        xs_list = []
        success_list = []
        for i in range(num_n_o):
            xs_list += [xs_e[:, :, i], xs_o[:, :, i]]
            success_list += [success_e[:, :, i], success_o[:, :, i]]
        if num_n_e > num_n_o:
            xs_list.append(xs_e[:, :, -1])
            success_list.append(success_e[:, :, -1])
        return np.dstack(xs_list), np.dstack(success_list)

    def betas_convs_samples(self, params: dict) -> tuple[dict, dict, Sampling]:
        im_factor = self.clad.im_factor
        self.clad.im_factor = 1.0
        self.clad_params["im_factor"] = 1.0
        p_modes = params["modes"].copy()
        num_n_0 = p_modes["num_n"]
        betas: dict = {}
        convs: dict = {}
        success = False
        catalog = Database().load_catalog()
        num_n_max = catalog["num_n"].max()
        if not np.isnan(num_n_max):
            for num_n in [n for n in range(num_n_0, num_n_max + 1, 2)]:
                p_modes["num_n"] = num_n
                smp = Samples(self.r, self.fill_params, self.clad_params, p_modes)
                try:
                    betas, convs = smp.database.load()
                    success = True
                    break
                except IndexError:
                    continue
        if not success:
            p_modes["num_n"] = num_n_0
            betas, convs, smp = self.do_sampling(p_modes)
        if im_factor != 1.0:
            self.clad.im_factor = im_factor
            self.clad_params["im_factor"] = im_factor
            betas, convs, smp = self.do_sampling_for_im_factor(betas, convs, p_modes)
        return betas, convs, smp

    def do_sampling(self, p_modes: dict) -> tuple[dict, dict, Samples]:
        num_n_0 = p_modes["num_n"]
        smp = Samples(self.r, self.fill_params, self.clad_params, p_modes)
        ray.shutdown()
        try:
            ray.init()
            p_modes_id = ray.put(p_modes)
            pool = ray.util.ActorPool(
                SamplesForRay.remote(
                    self.r, self.fill_params, self.clad_params, p_modes_id
                )
                for _ in range(psutil.cpu_count())
            )
            args0 = [
                ("M", "even", num_n_0),
                ("M", "odd", num_n_0),
                ("E", "even", num_n_0),
                ("E", "odd", num_n_0),
            ]
            xs_success_wr_list = list(
                pool.map(lambda a, arg: a.wr_sampling.remote(arg), args0)
            )
            num_wr = xs_success_wr_list[0][0].shape[0]
            args1 = []
            for i, (pol, parity, _) in enumerate(args0):
                xs_array, _ = xs_success_wr_list[i]
                for iwr in range(num_wr):
                    args1.append((pol, parity, num_n_0, iwr, xs_array[iwr]))
            xs_success_wi_list: list[tuple[np.ndarray, np.ndarray]] = list(
                pool.map(lambda a, arg: a.wi_sampling.remote(arg), args1)
            )
        finally:
            ray.shutdown()
        num_wi = xs_success_wi_list[0][0].shape[0]
        xs_success_list: list[tuple[np.ndarray, np.ndarray]] = []
        for j, (pol, parity, _) in enumerate(args0):
            num = xs_success_wi_list[j][0].shape[1]
            xs_array = np.zeros((num_wr, num_wi, num), dtype=complex)
            success_array = np.zeros((num_wr, num_wi, num), dtype=bool)
            for iwr in range(num_wr):
                i = j * num_wr + iwr
                xs_i, success_i = xs_success_wi_list[i]
                xs_array[iwr] = xs_i
                success_array[iwr] = success_i
            xs_success_list.append((xs_array, success_array))
        xs_success_M = self.merge_even_and_odd_data(
            xs_success_list[0], xs_success_list[1]
        )
        xs_success_E = self.merge_even_and_odd_data(
            xs_success_list[2], xs_success_list[3]
        )
        betas, convs = smp.betas_convs([xs_success_M, xs_success_E])
        smp.database.save(betas, convs)
        return betas, convs, smp

    def do_sampling_for_im_factor(
        self, betas: dict, convs: dict, p_modes: dict
    ) -> tuple[dict, dict, SamplesLowLoss]:
        smp = SamplesLowLoss(self.r, self.fill_params, self.clad_params, p_modes)
        try:
            betas, convs = smp.database.load()
        except IndexError:
            num_n = p_modes["num_n"]
            ns = list(range(num_n))
            ns_e = ns[::2]
            ns_o = ns[1::2]
            args = []
            for iwr in range(len(smp.ws)):
                for iwi in range(len(smp.wis)):
                    xis_list = [
                        [betas[("M", n, 1)][iwr, iwi] ** 2 for n in ns_e],
                        [betas[("M", n, 1)][iwr, iwi] ** 2 for n in ns_o],
                        [betas[("E", n, 1)][iwr, iwi] ** 2 for n in ns_e],
                        [betas[("E", n, 1)][iwr, iwi] ** 2 for n in ns_o],
                    ]
                    args.append((iwr, iwi, xis_list))
            ray.shutdown()
            try:
                ray.init()
                p_modes_id = ray.put(p_modes)
                pool = ray.util.ActorPool(
                    SamplesLowLossForRay.remote(
                        self.r, self.fill_params, self.clad_params, p_modes_id
                    )
                    for _ in range(psutil.cpu_count())
                )
                xs_success_list = list(
                    pool.map(lambda a, arg: a.task.remote(arg), args)
                )

            finally:
                ray.shutdown()
            betas, convs = smp.betas_convs(xs_success_list)
            smp.database.save(betas, convs)
        return betas, convs, smp

    def beta(self, w: complex, alpha: tuple[str, int, int]) -> complex:
        """Return phase constant

        Args:
            w: A complex indicating the angular frequency
            alpha: A tuple (pol, n, m) where pol is 'M' for TM mode or
                'E' for TE mode, n is the order of the mode, and m is
                the number of modes in the order and the polarization,
                which is always 1 in the slit case.
        Returns:
            h: A complex indicating the phase constant.
        """
        if self.clad.label == "PEC":
            return self.beta_pec(w, alpha)
        wr = w.real
        wi = w.imag
        hr = self.beta_funcs[(alpha, "real")](wr, wi)[0, 0]
        hi = self.beta_funcs[(alpha, "imag")](wr, wi)[0, 0]
        # if hr < 0:
        #     hr = 1e-16
        # if hi < 0:
        #     hi = 1e-16
        return hr + 1j * hi

    def beta_pec(self, w: complex, alpha: tuple[str, int, int]) -> complex:
        """Return phase constant of PEC waveguide

        Args:
            w: A complex indicating the angular frequency
            alpha: A tuple (pol, n, m) where pol is 'M' for TM mode or
                'E' for TE mode, n is the order of the mode, and m is the
                number of modes in the order and the polarization,
                which is always 1 in the slit case.
                Here, the order of TE mode starts with 1.
        Returns:
            h: A complex indicating the phase constant.
        """
        w_comp = w.real + 1j * w.imag
        pol, n, m = alpha
        val = cmath.sqrt(self.fill(w_comp) * w_comp ** 2 - (n * np.pi / self.r) ** 2)
        if abs(val.real) > abs(val.imag):
            if val.real < 0:
                val *= -1
        else:
            if val.imag < 0:
                val *= -1
        return val

    def coef(self, h, w, alpha):
        """Return the coefficients of TE- and TM- components which compose
        the hybrid mode.

        Args:
            h: A complex indicating the phase constant.
            w: A complex indicating the angular frequency
            alpha: A tuple (pol, n, m) where pol is 'M' for TM mode or
                'E' for TE mode, n is the order of the mode, and m is
                the number of modes in the order and the polarization,
                which is always 1 in the slit case.
        Returns:
            a: A complex indicating the coefficient of TE-component
            b: A complex indicating the coefficient of TM-component
        """
        pol, n, m = alpha
        w = w.real + 1j * w.imag
        h = h.real + 1j * h.imag
        if pol == "E":
            norm = self.norm(w, h, alpha, 1.0 + 0.0j, 0.0j)
            ai, bi = 1.0 / norm, 0.0
        else:
            norm = self.norm(w, h, alpha, 0.0j, 1.0 + 0.0j)
            ai, bi = 0.0, 1.0 / norm
        return ai, bi

    @staticmethod
    def sinc(x):
        x1 = x / np.pi
        return np.sinc(x1)

    def norm(self, w, h, alpha, a, b):
        a2_b2 = a ** 2 + b ** 2
        e1 = self.fill(w)
        e2 = self.clad(w)
        pol, n, m = alpha
        if self.clad(w).real < -1e6:
            if pol == "M" and n == 0:
                return cmath.sqrt(a2_b2 * self.r)
            else:
                return cmath.sqrt(a2_b2 * self.r / 2)
        u = self.samples.u(h ** 2, w, e1)
        # uc = u.conjugate()
        v = self.samples.v(h ** 2, w, e2)
        # vc = v.conjugate()
        if n % 2 == 0:
            if pol == "E":
                b_a = cmath.sin(u)
                parity = -1
            else:
                b_a = u / v * cmath.sin(u)
                parity = 1
        else:
            if pol == "E":
                b_a = cmath.cos(u)
                parity = 1
            else:
                b_a = -u / v * cmath.cos(u)
                parity = -1
        val = cmath.sqrt(
            a2_b2
            * self.r
            * (b_a ** 2 / (2 * v) + (1.0 + parity * self.sinc(2 * u)) / 2)
        )
        return val

    def Y(self, w, h, alpha, a, b):
        """Return the effective admittance of the waveguide mode

        Args:
            w: A complex indicating the angular frequency
            h: A complex indicating the phase constant.
            alpha: A tuple (pol, n, m) where pol is 'M' for TM mode or
                'E' for TE mode, n is the order of the mode, and m is
                the number of modes in the order and the polarization,
                which is always 1 in the slit case.
            a: A complex indicating the coefficient of TE-component
            b: A complex indicating the coefficient of TM-component
        Returns:
            y: A complex indicating the effective admittance
        """
        pol, n, m = alpha
        e1 = self.fill(w)
        e2 = self.clad(w)
        y_te = self.y_te(w, h)
        y_tm_in = self.y_tm_inner(w, h)
        y_tm_out = self.y_tm_outer(w, h)
        if e2.real < -1e6:
            if pol == "E":
                return y_te
            else:
                return y_tm_in
        u = self.samples.u(h ** 2, w, e1)
        v = self.samples.v(h ** 2, w, e2)
        if pol == "E":
            y_in = y_out = y_te
        else:
            y_in = y_tm_in
            y_out = y_tm_out
        if n % 2 == 0:
            if pol == "E":
                b_a = np.sin(u)
                parity = -1
            else:
                b_a = u / v * np.sin(u)
                parity = 1
        else:
            if pol == "E":
                b_a = np.cos(u)
                parity = 1
            else:
                b_a = -u / v * np.cos(u)
                parity = -1
        val = (
            (a ** 2 + b ** 2)
            * self.r
            * (
                y_out * b_a ** 2 / (2 * v)
                + (1.0 + parity * self.sinc(2 * u)) * y_in / 2
            )
        )
        return val

    def Yab(self, w, h1, s1, l1, n1, m1, a1, b1, h2, s2, l2, n2, m2, a2, b2):
        """Return the admittance matrix element of the waveguide modes

        Args:
            w: A complex indicating the angular frequency
            h1, h2: A complex indicating the phase constant.
            s1, s2: 0 for TE mode or 1 for TM mode
            l1, l2: 0 for h mode or 1 for v mode. In the slit case,
                l=h for TM mode, and l=v for TE mode.
            n1, n2: the order of the mode
            m1, m2: the number of modes in the order and the polarization.
                They are always 1 in the slit case.
            a1, a2: A complex indicating the coefficient of TE-component
            b1, b2: A complex indicating the coefficient of TM-component
        Returns:
            y: A complex indicating the effective admittance
        """
        if s1 != s2:
            return 0.0
        if n1 % 2 != n2 % 2:
            return 0.0
        e1 = self.fill(w)
        e2 = self.clad(w)
        y_te = self.y_te(w, h2)
        y_tm_in = self.y_tm_inner(w, h2)
        y_tm_out = self.y_tm_outer(w, h2)
        if e2.real < -1e6:
            if n1 != n2:
                return 0.0
            if s1 == 0:
                return y_te
            else:
                return y_tm_in
        ac = a1
        a = a2
        bc = b1
        b = b2
        uc = self.samples.u(h1 ** 2, w, e1)
        u = self.samples.u(h2 ** 2, w, e1)
        vc = self.samples.v(h1 ** 2, w, e2)
        v = self.samples.v(h2 ** 2, w, e2)
        if s1 == 0:
            y_in = y_out = y_te
            val = ac * a * self.r
        else:
            y_in = y_tm_in
            y_out = y_tm_out
            val = bc * b * self.r
        if n1 % 2 == 0:
            if s1 == 0:
                b_ac = np.sin(uc)
                b_a = np.sin(u)
                parity = -1
            else:
                b_ac = uc / vc * np.sin(uc)
                b_a = u / v * np.sin(u)
                parity = 1
        else:
            if s1 == 0:
                b_ac = np.cos(uc)
                b_a = np.cos(u)
                parity = 1
            else:
                b_ac = -uc / vc * np.cos(uc)
                b_a = -u / v * np.cos(u)
                parity = -1
        val *= (
            y_out * b_ac * b_a / (v + vc)
            + y_in * (self.sinc(u - uc) + parity * self.sinc(u + uc)) / 2
        )
        return val

    @staticmethod
    def y_te(w, h):
        return h / w

    def y_tm_inner(self, w, h):
        e = self.fill(w)
        return e * w / h

    def y_tm_outer(self, w, h):
        e = self.clad(w)
        return e * w / h

    def fields(self, x, y, w, dir, alpha, h, coef):
        """Return the electromagnetic field vectors for the specified mode and
        point

        Args:
            x: A float indicating the x coordinate [um]
            y: A float indicating the y coordinate [um]
            w: A complex indicating the angular frequency
            dir: "h" (horizontal polarization) or "v" (vertical polarization).
                In the slit case, dir='h' for TM and dir='v' for TE.
            alpha: A tuple (pol, n, m) where pol is 'M' for TM-like mode or
                'E' for TE-like mode, n is the order of the mode, and m is
                the number of modes in the order and the polarization,
                which is always 1 in the slit case.
            h: A complex indicating the phase constant.
            coef: The coefficients of TE- and TM- components
        Returns:
            f_vec: An array of complexes [ex, ey, ez, hx, hy, hz].
        """
        pol, n, m = alpha
        a, b = coef
        u = self.samples.u(h ** 2, w, self.fill(w))
        v = self.samples.v(h ** 2, w, self.clad(w))
        gd = u / (self.r / 2)
        gm = v / (self.r / 2)
        if pol == "E":
            y_te = self.y_te(w, h)
            ex = ez = 0.0
            hy = 0.0
            if n % 2 == 1:
                # parity even
                if abs(x) <= self.r / 2:
                    ey = a * np.cos(gd * x)
                    hx = y_te * ey
                    hz = 1j * gd / w * a * np.sin(gd * x)
                else:
                    b_a = np.exp(v) * np.cos(u)
                    ey = a * b_a * np.exp(-gm * abs(x))
                    hx = y_te * ey
                    hz = 1j * gm / w * x / abs(x) * ey
            else:
                # parity odd
                if abs(x) <= self.r / 2:
                    ey = a * np.sin(gd * x)
                    hx = y_te * ey
                    hz = -1j * gd / w * a * np.cos(gd * x)
                else:
                    b_a = np.exp(v) * np.sin(u)
                    ey = a * b_a * x / abs(x) * np.exp(-gm * abs(x))
                    hx = y_te * ey
                    hz = 1j * gm / w * x / abs(x) * ey
        else:
            hx = hz = 0.0
            ey = 0.0
            if n % 2 == 0:
                # parity even
                if abs(x) <= self.r / 2:
                    y_tm = self.y_tm_inner(w, h)
                    ex = b * np.cos(gd * x)
                    hy = y_tm * ex
                    ez = -1j * gd / h * b * np.sin(gd * x)
                else:
                    y_tm = self.y_tm_outer(w, h)
                    b_a = u / v * np.exp(v) * np.sin(u)
                    ex = b * b_a * np.exp(-gm * abs(x))
                    hy = y_tm * ex
                    ez = -1j * gm * x / abs(x) / h * ex
            else:
                # parity odd
                if abs(x) <= self.r / 2:
                    y_tm = self.y_tm_inner(w, h)
                    ex = b * np.sin(gd * x)
                    hy = y_tm * ex
                    ez = 1j * gd / h * b * np.cos(gd * x)
                else:
                    y_tm = self.y_tm_outer(w, h)
                    b_a = -u / v * np.exp(v) * np.cos(u)
                    ex = b * b_a * x / abs(x) * np.exp(-gm * abs(x))
                    hy = y_tm * ex
                    ez = -1j * gm * x / abs(x) / h * ex
        return np.array([ex, ey, ez, hx, hy, hz])

    def e_field(self, x, y, w, dir, alpha, h, coef) -> np.ndarray:
        """Return the electric field vector for the specified mode and
        point

        Args:
            x: A float indicating the x coordinate [um]
            y: A float indicating the y coordinate [um]
            w: A complex indicating the angular frequency
            dir: "h" (horizontal polarization) or "v" (vertical polarization).
                In the slit case, dir='h' for TM and dir='v' for TE.
            alpha: A tuple (pol, n, m) where pol is 'M' for TM mode or
                'E' for TE mode, n is the order of the mode, and m is
                the number of modes in the order and the polarization,
                which is always 1 in the slit case.
            h: A complex indicating the phase constant.
            coef: The coefficients of TE- and TM- components
        Returns:
            e_vec: Array(ex, ey, ez).
        """
        pol, n, m = alpha
        a, b = coef
        u = self.samples.u(h ** 2, w, self.fill(w))
        v = self.samples.v(h ** 2, w, self.clad(w))
        gd = u / (self.r / 2)
        gm = v / (self.r / 2)
        if pol == "E":
            ex = ez = 0.0
            if n % 2 == 1:
                # parity even
                if abs(x) <= self.r / 2:
                    ey = a * np.cos(gd * x)
                else:
                    b_a = np.exp(v) * np.cos(u)
                    ey = a * b_a * np.exp(-gm * abs(x))
            else:
                # parity odd
                if abs(x) <= self.r / 2:
                    ey = a * np.sin(gd * x)
                else:
                    b_a = np.exp(v) * np.sin(u)
                    ey = a * b_a * x / abs(x) * np.exp(-gm * abs(x))
        else:
            ey = 0.0
            if n % 2 == 0:
                # parity even
                if abs(x) <= self.r / 2:
                    ex = b * np.cos(gd * x)
                    ez = -1j * gd / h * b * np.sin(gd * x)
                else:
                    b_a = u / v * np.exp(v) * np.sin(u)
                    ex = b * b_a * np.exp(-gm * abs(x))
                    ez = -1j * gm * x / abs(x) / h * b * b_a * np.exp(-gm * abs(x))
            else:
                # parity odd
                if abs(x) <= self.r / 2:
                    ex = b * np.sin(gd * x)
                    ez = 1j * gd / h * b * np.cos(gd * x)
                else:
                    b_a = -u / v * np.exp(v) * np.cos(u)
                    ex = b * b_a * x / abs(x) * np.exp(-gm * abs(x))
                    ez = -1j * gm / h * b * b_a * np.exp(-gm * abs(x))
        return np.array([ex, ey, ez])

    def h_field(self, x, y, w, dir, alpha, h, coef) -> np.ndarray:
        """Return the magnetic field vectors for the specified mode and
        point

        Args:
            x: A float indicating the x coordinate [um]
            y: A float indicating the y coordinate [um]
            w: A complex indicating the angular frequency
            dir: "h" (horizontal polarization) or "v" (vertical polarization).
                In the slit case, dir='h' for TM and dir='v' for TE.
            alpha: A tuple (pol, n, m) where pol is 'M' for TM mode or
                'E' for TE mode, n is the order of the mode, and m is
                the number of modes in the order and the polarization,
                which is always 1 in the slit case.
            h: A complex indicating the phase constant.
            coef: The coefficients of TE- and TM- components
        Returns:
            h_vec: Array(hx, hy, hz).
        """
        pol, n, m = alpha
        a, b = coef
        u = self.samples.u(h ** 2, w, self.fill(w))
        v = self.samples.v(h ** 2, w, self.clad(w))
        gd = u / (self.r / 2)
        gm = v / (self.r / 2)
        if pol == "E":
            y_te = self.y_te(w, h)
            hy = 0.0
            if n % 2 == 1:
                # parity even
                if abs(x) <= self.r / 2:
                    hx = y_te * a * np.cos(gd * x)
                    hz = 1j * gd / w * a * np.sin(gd * x)
                else:
                    b_a = np.exp(v) * np.cos(u)
                    hx = y_te * a * b_a * np.exp(-gm * abs(x))
                    hz = 1j * gm / w * x / abs(x) * a * b_a * np.exp(-gm * abs(x))
            else:
                # parity odd
                if abs(x) <= self.r / 2:
                    hx = y_te * a * np.sin(gd * x)
                    hz = -1j * gd / w * a * np.cos(gd * x)
                else:
                    b_a = np.exp(v) * np.sin(u)
                    hx = y_te * a * b_a * x / abs(x) * np.exp(-gm * abs(x))
                    hz = (
                        1j
                        * gm
                        / w
                        * x
                        / abs(x)
                        * a
                        * b_a
                        * x
                        / abs(x)
                        * np.exp(-gm * abs(x))
                    )
        else:
            hx = hz = 0.0
            if n % 2 == 0:
                # parity even
                if abs(x) <= self.r / 2:
                    y_tm = self.y_tm_inner(w, h)
                    hy = y_tm * b * np.cos(gd * x)
                else:
                    y_tm = self.y_tm_outer(w, h)
                    b_a = u / v * np.exp(v) * np.sin(u)
                    hy = y_tm * b * b_a * np.exp(-gm * abs(x))
            else:
                # parity odd
                if abs(x) <= self.r / 2:
                    y_tm = self.y_tm_inner(w, h)
                    hy = y_tm * b * np.sin(gd * x)
                else:
                    y_tm = self.y_tm_outer(w, h)
                    b_a = -u / v * np.exp(v) * np.cos(u)
                    hy = y_tm * b * b_a * x / abs(x) * np.exp(-gm * abs(x))
        return np.array([hx, hy, hz])

    def coefs_numpy(self, hs, w):
        As = []
        Bs = []
        for h, s, n, m in zip(hs, self.s_all, self.n_all, self.m_all):
            pol = "E" if s == 0 else "M"
            ai, bi = self.coef(h, w, (pol, n, m))
            As.append(ai)
            Bs.append(bi)
        return np.ascontiguousarray(As), np.ascontiguousarray(Bs)

    def coefs(self, hs, w):
        return slit_utils.coefs_cython(self, hs, w)

    def Ys(self, w, hs, As, Bs):
        vals = []
        for h, s, n, a, b in zip(hs, self.s_all, self.n_all, As, Bs):
            pol = "E" if s == 0 else "M"
            vals.append(self.Y(w, h, (pol, n, 1), a, b))
        return np.array(vals)

    def hAB(self, w):
        hs = np.array([self.beta(w, alpha) for alpha in self.alpha_all])
        As, Bs = self.coefs(hs, w)
        return hs, As, Bs

    def ABY(self, w, hs):
        e1 = self.fill(w)
        e2 = self.clad(w)
        return slit_utils.ABY_cython(w, self.r, self.s_all, self.n_all, hs, e1, e2)

    def hABY(self, w):
        e1 = self.fill(w)
        e2 = self.clad(w)
        hs = np.array([self.beta(w, alpha) for alpha in self.alpha_all])
        As, Bs, Y = slit_utils.ABY_cython(w, self.r, self.s_all, self.n_all, hs, e1, e2)
        return hs, As, Bs, Y

    def huvABY(self, w):
        e1 = self.fill(w)
        e2 = self.clad(w)
        hs = np.array([self.beta(w, alpha) for alpha in self.alpha_all])
        us, vs, As, Bs, Y = slit_utils.uvABY_cython(
            w, self.r, self.s_all, self.n_all, hs, e1, e2
        )
        return hs, us, vs, As, Bs, Y
